//	NEBApp.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI
import CompositorServices

//	Note:  If you go to
//
//		Target > Build Settings > Swift Compiler - Custom Flags > Other Swift Flags > Debug
//
//	and add the lines
//
//		-Xfrontend
//		-warn-long-function-bodies=100
//		-Xfrontend
//		-warn-long-expression-type-checking=5
//
//	the compiler will measure the time required for type-checking individual lines of code
//	and for compiling individual functions, and will issue warnings for slow ones.
//	Unfortunately there are many lines of code that are slow to type-check.
//	Note: Accessing a component of a vector v as v.x is faster to type check
//	than accessing it as v[0].

//	Should we color only the balls that the player
//	is currently trying to sink?
//
//		The public version of Curved-Space Pool Hall keeps
//		colorTargetBallsOnly enabled at all times.
//		The only reason I included this flag is so that
//		I can disable colorTargetBallsOnly while making
//		screenshots for use in lectures etc.
//
//		The file
//			scraps/colorTargetBallsOnly.swift
//		contains the old code that included a control panel
//		toggle to let the player enable or disable
//		colorTargetBallsOnly on the fly, along with
//		an Info panel explaining it.
//
//#warning("colorTargetBallsOnly is disabled -- reenable for public release")
let colorTargetBallsOnly: Bool = true

enum InstructionItem {
	case walkBackwards
	case handsAboveTable
	case placeCueBall
	case gripHandHeight
	case howToReturn
	case startPlaying
}

//	I'm a little embarrassed to define gInstructionsPanelHasClosed
//	as a global variable, but this seems like the simplest way
//	to let the Control Panel tell the Renderer that it's time
//	to give the player a cue stick. The seemingly more natural
//	approach of calling a method in the renderer doesn't work
//	because the renderer is running uninterrupted on its own Actor,
//	with no await calls, so once we call runRenderLoop() there's
//	no longer any way to get the renderer to accept calls to any
//	of its other methods.
//
//		Note #1: I've wrapped this Bool in an AtomicBool struct
//		that protects all reads and writes with an NSLock, but it
//		would probably also be OK to take the simpler approach of
//		a plain unprotected Bool
//
//			nonisolated(unsafe) var gInstructionsPanelHasClosed: Bool = false
//
//		Reading and writing Bools might well be completely
//		atomic operations, and even if they aren't, it's hard to imagine
//		that while one thread is replacing 0x00000000 with 0x00000001,
//		that the other thread could read any value other than
//		0x00000000 or 0x00000001.
//
//		Note #2: Adam says that Swift Atomics probably contains
//		native atomic read/write operations, see
//			https://github.com/apple/swift-atomics
//		and various other web pages. However, installing a separate
//		package, and keeping it updated as this project
//		goes forward, seems a cure that's worse than the disease.
//		But if Swift Atomics ever get included in the Swift Standard Library,
//		I could use a natively atomic Bool to replace my ad hoc AtomicBool.
//
//		Note #3: In a purely SwiftUI context, a Binding to a @State variable
//		would be the right solution. But our NEBRenderer is a plain Actor,
//		with no connection to SwiftUI at all. It would be possible
//		to import the appropriate SwiftUI header and used a Binding<>
//		even in NEBRenderer, but I'd rather not introduce that extra
//		complication. Also, the NEBRenderer runs in its own executor,
//		and who know what delicate interactions we'd need to deal with
//		if a Binding in one executor were fetching information from
//		a @State variable in another executor. Better to stick
//		with a good ol' NSLock(). The lock will essentially never
//		be contested, so it will get locked and unlocked very quickly
//		and reliable, all on the same executor running on the same thread.
//
struct AtomicBool {
	private var value: Bool
	private var itsLock = NSLock()
	init(_ initialValue: Bool) {
		value = initialValue
	}
	func get() -> Bool {
		itsLock.lock()
		let theValue = value
		itsLock.unlock()
		return theValue
	}
	mutating func set(_ newValue: Bool) {
		itsLock.lock()
		value = newValue
		itsLock.unlock()
	}
}
nonisolated(unsafe) var gInstructionsPanelHasClosed = AtomicBool(false)

//	At a lecture or other public event, having each person
//	do the eye calibration procedure wastes a huge amount of time.
//	By making the game selection button's huge, people may be able
//	to skip the calibration and play with mis-calibrated vision.
//	In case of problem, I can quadruple-press the Top Button and
//	work the Digital Crown for them to do the hand & eye setup
//	without having to go through all the other hassle of Guest Mode.
//
//	When people use the app only briefly at public events,
//	we'll also want to strengthen the Guardian Angel,
//	weaken the computer player, make the table a little higher,
//	and suppress the optional hints.
//
//		Caution:   We want to leave all settings fully adjustable
//		for real-time flexibility. So setting different default values
//		doesn't quite guarantee that the actual values won't get
//		unadvertantly altered before the lecture begins.
//		I've left myself a reminder in my lecture notes (in Keynote)
//		to double-check all the desired values at some point
//		before the talk (they persist in User Prefs).
//
//#warning("Be sure forTalks is false when submitting to the App Store.")
let forTalks = false


@main
struct NEBApp: App {

	@AppStorage("Player handedness") var playerHandedness: PlayerHandedness = .rightHanded

	//	How well should the computer play?
	@AppStorage("Difficulty") var computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
									 = forTalks ? 0.25 : 0.5
	
	//	Especially at public events where each player might have only
	//	a couple minutes to try the app, novices might appreciate
	//	a helping hand to make their shots more successful.
	@AppStorage("Guardian angel factor") var guardianAngelFactor: Double	//	∈ 0.0 ... 1.0
									= forTalks ? 1.0 : 0.75
	
	@State var geometryType: GeometryType = .euclidean

	@AppStorage("Table width") var tableEdgeLengthInMeters: Double = defaultTableEdgeLengthInMeters
	@AppStorage("Table height") var tableHeight: Double = defaultTableHeight	//	in meters

	//	Should the renderer adjust the distances so that
	//	a Euclidean born-and-raised tourist will see each object
	//	at the same distance that a native-born inhabitant
	//	of the non-Euclidean space would see it?
	//
	//	For a full discussion of tourist view vs. native view, see
	//
	//		Jeff Weeks,
	//		Non-Euclidean Billiards in VR, Proceedings of Bridges 2020:
	//		Mathematics, Music, Art, Architecture, Culture (2020) 1–8.
	//
	//		https://archive.bridgesmathart.org/2020/bridges2020-1.pdf
	//
	//	and
	//
	//		Jeff Weeks,
	//		Body coherence in curved-space virtual reality games,
	//		Computers & Graphics 97 (2021) 28–41.
	//
	//		  https://www.sciencedirect.com/science/article/pii/S0097849321000443
	//		or
	//		  https://arxiv.org/abs/2011.00510
	//
	//	2024-10-03  Now that we no longer anchor the player's hands
	//	to a physical table, but instead anchor them to the player's head
	//	and map them into the space using the "exponential map",
	//	it makes sense to set adjustPerceivedDistances to true by default,
	//	so the player will see his/her hands exactly where s/he feels them to be.
	//
	@AppStorage("Adjust perceived distances") var adjustPerceivedDistances: Bool = true
	
	//	The user may set "Don't show hints again" to disable
	//	the optional hints panels and then later go to the Control Panel
	//	to re-enable them via "Show hints when starting game".
	@AppStorage("Show hints") var showHints: Bool = forTalks ? false : true

	//	Normally one would expect the instructionsViewIsPresented
	//	and the instructionItem to @State variables within
	//	the NEBControlPanel and the InstructionsView, respectively,
	//	but we need to define them here at the NEBApp level so that
	//	the closures we pass to the NEBRenderer can set those
	//	variables appropriately when when the player steps backwards,
	//	places the cue ball on the table, or closes the immersive view.
	@State var instructionsViewIsPresented: Bool = false
	@State var instructionItem: InstructionItem = .walkBackwards

	//	Apple's .upperLimbVisibility() page
	//
	//		https://developer.apple.com/documentation/swiftui/view/upperlimbvisibility(_:)
	//
	//	says
	//
	//		Note that this modifier only sets a preference and ultimately
	//		the system will decide if it will honor it or not. The system
	//		may be unable to honor the preference if the immersive space
	//		is currently not visible.
	//
	//	My first reaction was, how could the immersive space already be visible,
	//	if we supply the .upperLimbVisibility() modifier as part of
	//	our request to create it? The resolution of this paradox is that
	//	visionOS uses the (transient!) SwiftUI ImmersiveSpace struct
	//	to create an immersive space in some private internal format.
	//	If we specify the upper-limb visibility as
	//
	//		.upperLimbVisibility(hideHands ? .hidden : .visible)
	//
	//	we can switch from visible hands to hidden hands at some time
	//	after the immersive view is already visible. In particular,
	//	changing the parameter passed to .upperLimbVisibility()
	//	doesn't cause visionOS to re-create the immersive view.
	//	It keeps the immersive view that's already visible and
	//	just changes the upper-limb visibility. Whew!
	@State var hideHands: Bool = false

	@Environment(\.openWindow) var openWindow
	@Environment(\.dismissWindow) var dismissWindow


	init() {

		//	Let the NEBSoundPlayer load its sounds
		//	well before the player takes his/her first shot.
		Task { await NEBSoundPlayer.shared.hello() }
	}

	var body: some Scene {

		WindowGroup(id: "Control panel", for: Int.self) { $foo in
		
			NEBControlPanel(
			playerHandedness: $playerHandedness,
			computerPlayerDifficulty: $computerPlayerDifficulty,
			guardianAngelFactor: $guardianAngelFactor,
			geometryType: $geometryType,
			tableEdgeLengthInMeters: $tableEdgeLengthInMeters,
			tableHeight: $tableHeight,
			adjustPerceivedDistances: $adjustPerceivedDistances,
			showHints: $showHints,
			instructionsViewIsPresented: $instructionsViewIsPresented,
			instructionItem: $instructionItem,
			hideHands: $hideHands)
			
		} defaultValue: { 0 }	//	assigns value 0 to the initial window
		.defaultSize(
			width: 256.0,
			height: forTalks ? 1280.0 : 288.0)

		ImmersiveSpace(id: "Billiards immersive space") {

			CompositorLayer(configuration: GGVRCompositorLayerConfiguration()
			) { layerRenderer in
			
				let thePlayerSteppedBack = {  @Sendable () -> Void in
					Task { @MainActor in
						instructionItem = .handsAboveTable
					}
				}
				let thePlayerRaisedCueBall = {  @Sendable () -> Void in
					Task { @MainActor in
						instructionItem = .placeCueBall
					}
				}
				let thePlayerPlacedCueBall = {  @Sendable () -> Void in
					Task { @MainActor in
						if showHints {
							instructionItem = .gripHandHeight
						} else {
							instructionsViewIsPresented = false
							gInstructionsPanelHasClosed.set(true)
							dismissWindow(id: "Control panel")
						}
					}
				}
				let thePlayerClosedImmersiveSpace = {  @Sendable () -> Void in
					Task { @MainActor in
					
						//	The Control Panel may or may not be up already.
						//	Either way, we set instructionsViewIsPresented to false
						//	and gInstructionsPanelHasClosed to true.
						instructionsViewIsPresented = false
						gInstructionsPanelHasClosed.set(true)

						//	openWindow(id: …, value: 0) uses the existing
						//	Control Panel if one is up.
						openWindow(id: "Control panel", value: 0)
					}
				}

#warning("Delete theTemporaryClosure in some future release, when players are used to pressing the Digital Crown to close the immersive space.")
				let theTemporaryClosure = {  @Sendable () -> Void in
					Task { @MainActor in
						openWindow(id: "TEMPORARY alert", value: 0)
					}
				}

#warning("layerRenderer isn't sendable, so when Strict Concurrency Checking is set to 'complete', the compiler complains here. Xcode's template code has this same problem.")
				Task(executorPreference: GGVRRendererTaskExecutor.shared) {
				
					guard let theRenderer = await NEBRenderer(
							layerRenderer: layerRenderer,
							playerHandedness: playerHandedness,
							computerPlayerDifficulty: computerPlayerDifficulty,
							guardianAngelFactor: guardianAngelFactor,
							adjustPerceivedDistances: adjustPerceivedDistances,
							geometryType: geometryType,
							tableEdgeLengthInMeters: tableEdgeLengthInMeters,
							tableHeight: tableHeight,
							playerSteppedBack: thePlayerSteppedBack,
							playerRaisedCueBall: thePlayerRaisedCueBall,
							playerPlacedCueBall: thePlayerPlacedCueBall,
							playerClosedImmersiveSpace: thePlayerClosedImmersiveSpace,
							temporaryClosure: theTemporaryClosure)
					else {
						assertionFailure("Failed to create theRenderer")
						return
					}

					await theRenderer.startWorldAndHandTracking()
					await theRenderer.runRenderLoop()
					
					//	Note: runRenderLoop() contains no 'await' calls
					//	and runs without interruption, which is just
					//	what we want for a render loop.
					//	But be aware that if some other Task
					//	tries to call one of the renderer's methods,
					//	the call will never be received.
				}
			}
		}
		.immersionStyle(selection: .constant(.full), in: .full)
		.upperLimbVisibility(hideHands ? .hidden : .visible)
		.persistentSystemOverlays(.hidden)
			//	Asks visionOS not to interpret palm-up as a request for a home button.
			//	visionOS may, at whim, choose not to comply.

#warning("Delete this temporary alert in some future release, when players are used to pressing the Digital Crown to close the immersive space.")
		WindowGroup(id: "TEMPORARY alert", for: Int.self) { _ in
			TEMPORARYView()
		} defaultValue: { 0 }	//	assigns value 0 to the initial window
		.defaultSize(width: 256.0, height: 288.0)
	}
}

#warning("Delete this TEMPORARYView in some future release, when players are used to pressing the Digital Crown to close the immersive space.")
struct TEMPORARYView: View {

	@Environment(\.dismissWindow) var dismissWindow

	var body: some View {

		VStack(spacing: 16.0) {

			Image(systemName: "digitalcrown.horizontal.press")
			.resizable()
			.scaledToFit()
			.frame(width: 64.0, height: 64.0)

			Text("To return to the control panel,\nplease press the Digital Crown.")
			.fixedSize(horizontal: false, vertical: true)
				//	fixedSize() is needed to allow multiple lines
				//	if the enclosing frame's height isn't specified
				//	(unintuitive but true)
			.multilineTextAlignment(.center)

			Text("You no longer need to raise\nyour hand for this purpose.")
			.font(.caption2)
			.fixedSize(horizontal: false, vertical: true)
				//	fixedSize() is needed to allow multiple lines
				//	if the enclosing frame's height isn't specified
				//	(unintuitive but true)
			.multilineTextAlignment(.center)

			Button("OK") {
				dismissWindow(id: "TEMPORARY alert", value: 0)
			}
		}
	}
}
