//	NEBComputerPlayer.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import Foundation

//	How long should the cue stick sit still before and after the stroke?
//
//		Note:  The human player may see a computer opponent's
//		cue stick fade in when the computer opponent gets the turn,
//		but if the computer opponent takes additional shots as part
//		of the same turn, the transitions will be abrupt.
//
let preStrokeInterval			//	in seconds
	= 1.0 + max(colorSaturationChangeDuration, cueStickOpacityChangeDuration)
let postStrokeInterval = 1.0	//	in seconds

//	How much error should the easiest difficulty setting introduce?
let maxAngleStdDev = 0.1	//	radians
let maxLogSpeedStdDev = 0.2	//	dimensionless


class NEBComputerPlayer {
	
	private var shotSpaceGeometry: NEBSpaceGeometry?

	private var shotStartTime: Double?

	//	Pose at which the cue stick's tip reaches
	//	the cue ball's surface.  In table coordinates.
	private var poseAtImpact: simd_double4x4?
	
	//	The comments in planShot() explain these parameters.
	private var strokeHalfLength: Double?
	private var strokePeriod: Double?
	
	
	init() {
	}
	
	func shotIsInProgress(
	) -> Bool {
	
		return (shotSpaceGeometry != nil)
	}


// MARK: -
// MARK: Cue stick pose

	func cueStickPose(
		atTime time: TimeInterval	//	time since last reboot
	) -> simd_double4x4? {

		//	The comments in planShot() explain these parameters.
		if let theSpaceGeometry = shotSpaceGeometry,
		   let theShotStartTime = shotStartTime,
		   let thePoseAtImpact = poseAtImpact,
		   let d = strokeHalfLength,
		   let p = strokePeriod {

			let totalElapsedShotTime = time - theShotStartTime
			
			let t: Double
			if totalElapsedShotTime < preStrokeInterval {
			
				t = 0.0
				
			} else if totalElapsedShotTime < preStrokeInterval + p {
			
				t = totalElapsedShotTime - preStrokeInterval
				
			} else if totalElapsedShotTime < preStrokeInterval + p + postStrokeInterval {
			
				t = p	//	same effect as t = 0.0
				
			} else {
			
				//	The shot is over.

				shotSpaceGeometry = nil
				shotStartTime = nil
				poseAtImpact = nil
				strokeHalfLength = nil
				strokePeriod = nil
				
				return nil
			}

			//	The comments in planShot() explain the formula
			//
			//		z = d cos(2π t/p)
			//
			let z = d * cos(2.0 * Double.pi * t / p)	//	in meters
			
			let theOffset = axisAlignedTranslation(
				axis: 2,
				distanceInRadians: z / theSpaceGeometry.spaceRadius,
				geometryType: theSpaceGeometry.geometryType)
			
			let theCueStickPose = thePoseAtImpact * theOffset
									//	right-to-left matrix action

			return theCueStickPose
			
		} else {
		
			return nil	//	no shot in progress
		}
	}
	

// MARK: -
// MARK: Plan shot
	
	func planShot(
		in gameState: NEBGameState,
		shooterStatus: PlayerStatus,
		difficulty: Double,	//	∈ 0.0 (easy) ... 1.0 (hard)
		startTime: TimeInterval	//	time since last reboot
	) {
		//	Decide what shot to take.
		
		//	Note #1:  We don't modify or save the gameState.
		//	We examine it here to decide on our shot and then we forget it.
		//	We affect the game only through the values we return
		//	from cueStickPose().
		
		//	Note #2:  Because it considers only direct shots
		//	and ignores shots involving the cushions (for the
		//	reasons explained in detail in feasibleShots()),
		//	planShot() is fast. On my original Vision Pro
		//	it takes only a few tenths of a millisecond.
		//	So there's no problem running it synchronously
		//	on theRenderThread.
		
		precondition(
			gameState.balls[0].status == .inPlay,
			"planShot() called when cue ball is not in play")

		let theFeasibleShots = feasibleShots(
								in: gameState,
								shooterStatus: shooterStatus,
								guaranteeNonemptyResult: true)
		
		//	Select the ShotCandidate with the slowest startSpeed,
		//	because such shots tend to be simple and robust
		//	even after applying Gaussian error.
		let thePreciseShot = slowestShot(of: theFeasibleShots)

		let theImpreciseShot = applyGaussianError(
								to: thePreciseShot,
								difficulty: difficulty)

		requestShot(
			theImpreciseShot,
			in: gameState,
			startTime: startTime)
	}

	private func requestShot(
		_ requestedShot: ShotCandidate,
		in gameState: NEBGameState,
		startTime: TimeInterval	//	time since last reboot
	) {
	
		let thePoseAtBallCenter				//	in table coordinates
			= gameState.balls[0].placement	//	right-to-left matrix action
			* gameState.spaceGeometry.ballLift
		
		let theRotation = axisAlignedRotation(
			from: 0,
			towards: 2,
			angle: requestedShot.angle)

		let theOffsetAtImpact = axisAlignedTranslation(
			axis: 2,
			distanceInRadians: ballRadius / gameState.spaceGeometry.spaceRadius,
			geometryType: gameState.spaceGeometry.geometryType)

		let thePoseAtImpact			//	in table coordinates
			= thePoseAtBallCenter	//	right-to-left matrix action
			* theRotation
			* theOffsetAtImpact

		//	Keeping in mind that the z-axis points backwards,
		//	we can model the full forward and backward stroke as
		//
		//		z = d cos(2π t/p)
		//
		//	where d is the "half length" of stroke
		//	and p is the period (even though it's not really periodic
		//	here because we're doing only one cycle).
		//	z is measured relative to point of impact,
		//	so the cue tip strikes the cue ball at t = p/4.
		
		//	We can set d to whatever distance looks good,
		//	then compute the required period p to get the speed right.
		let d = 0.05	//	in meters
		
		let theCueTipSpeedAtImpact = requestedShot.startSpeed / cueStickBounceFactor
		
		//	The cue tip's speed at any time t is
		//
		//		dz
		//		-- = -d (2π/p) sin(2π t/p)     in meters/second
		//		dt
		//
		//	so at the moment of impact, when t = p/4, the speed is
		//
		//		dz
		//		-- = -d (2π/p) sin(2π (p/4)/p)
		//		dt
		//		   = -d (2π/p) sin(π/2)
		//		   = -d (2π/p)                 in meters/second
		//
		//	If we set
		//
		//		-cueTipSpeedAtImpact = -d (2π/p)
		//
		//	we can solve for the period
		//
		//		p = 2π d / cueTipSpeedAtImpact    in seconds
		//
		let p = 2.0 * Double.pi * d / theCueTipSpeedAtImpact

		shotSpaceGeometry = gameState.spaceGeometry
		shotStartTime = startTime
		poseAtImpact = thePoseAtImpact
		strokeHalfLength = d
		strokePeriod = p
	}
}

struct ShotCandidate {

	//	Rotation in the XZ coordinate plane
	//	(rotating the positive X axis towards the positive Z axis)
	//	in the cue ball's post-rotation coordinate system.
	//
	//	All our coordinate systems follow Metal's convention
	//	of letting the positive Z axis point towards the observer,
	//	so the cue ball's direction of motion is the negative Z axis.
	//	ShotCandidate.angle tells how to rotate that negative Z axis
	//	to the proposed direction for a shot.
	//
	let angle: Double		//	radians

	let startSpeed: Double	//	meters/second
}

func feasibleShots(
	in gameState: NEBGameState,
	shooterStatus: PlayerStatus,
	guaranteeNonemptyResult: Bool
) -> [ShotCandidate] {

	guard let theCueBallPose = getCueBallPoseXZW(in: gameState) else {
		preconditionFailure("feasibleShots() called when cue ball is not in play")
	}
	let theTargetBallCenters = getTargetBallCentersXZW(
								in: gameState,
								shooterStatus: shooterStatus)
	let theAllBallCenters = getAllBallCentersXZW(in: gameState)	//	excludes cue ball

	//	"Direct shots" are those in which the cue ball hits
	//	an object ball which then rolls directly into a pocket.
	//
	//	Curved-Space Pool Hall doesn't consider "indirect shots"
	//	in which the cue ball and/or the object ball bounces
	//	off of a cushion en route, because such shots would require
	//	more work to code up and would be much slower at run time.
	//	The reason they're more difficult is that the "orthogonal shadow"
	//	of a ball's center reflects not off of the pool table's
	//	geodesic edge (which would be easy to handle!), but rather
	//	off of a non-geodesic curve that runs "parallel" to the table edge
	//	but is displaced one ball radius inward toward the center
	//	of the table. Reflecting off of that non-geodesic curve
	//	would (I think) require some messy numerical method
	//	to find the required point of impact. I don't see how
	//	to compute the impact point in close form.
	//
	//	In any case, the ComputerPlayer plays plenty well enough
	//	using only direct shots.
	//
	let theDirectShots = directShots(
							in: gameState,
							cueBallPose: theCueBallPose,
							targetBallCenters: theTargetBallCenters,
							allBallCenters: theAllBallCenters)

	if guaranteeNonemptyResult
	&& theDirectShots.isEmpty {
	
		let theHailMaryShot = chooseHailMaryShot(
								in: gameState,
								cueBallPose: theCueBallPose,
								targetBallCenters: theTargetBallCenters)
	
		return [theHailMaryShot]
		
	} else {
	
		return theDirectShots
	}
}

private func getCueBallPoseXZW(
	in gameState: NEBGameState
) -> matrix_double3x3? {	//	nil if cue ball is not inPlay

	let theCueBall = gameState.balls[0]
	
	if theCueBall.status == .inPlay {

		let p = theCueBall.placement
		
		let theCueBallPoseXZW = matrix_double3x3(
			SIMD3<Double>(p[0][0], p[0][2], p[0][3]),	//	omits XY entry
			//	omits (YX, YY, YZ, YW) row)
			SIMD3<Double>(p[2][0], p[2][2], p[2][3]),	//	omits ZY entry
			SIMD3<Double>(p[3][0], p[3][2], p[3][3])	//	omits WY entry
		)
		
		return theCueBallPoseXZW

	} else {
		return nil
	}
}

private func getTargetBallCentersXZW(
	in gameState: NEBGameState,
	shooterStatus: PlayerStatus
) -> [VectorXZW] {

	let theTargetRange: Array<Int>
	switch shooterStatus {
	
	case .shootingForEither:
		theTargetRange = Array(1...7) + Array(9...15)
	case .shootingForSolids:
		theTargetRange = Array(1...7)
	case .shootingForStripes:
		theTargetRange = Array(9...15)
	case .shootingForEightBall:
		theTargetRange = [8]
	case .won, .lost:
		assertionFailure("getTargetBallCentersXZW() called when game is already over")
		theTargetRange = []
	}
	
	let theTargetBallCentersXZW = theTargetRange.compactMap() { i in
		
		getBallCenterXZW(
			ball: gameState.balls[i],
			spaceGeometry: gameState.spaceGeometry)
	}
	
	return theTargetBallCentersXZW
}

private func getAllBallCentersXZW(
	in gameState: NEBGameState
) -> [VectorXZW] {

	let theBallCentersXZW = Array(1...15).compactMap() { i in
		
		getBallCenterXZW(
			ball: gameState.balls[i],
			spaceGeometry: gameState.spaceGeometry)
	}
	
	return theBallCentersXZW
}

private func getBallCenterXZW(
	ball: BilliardBall,
	spaceGeometry: NEBSpaceGeometry
) -> VectorXZW? {
		
	if ball.status == .inPlay {
	
		let theTargetBallCenter = ball.placement * spaceGeometry.northPole

		let theTargetBallCenterXZW
			= VectorXZW(
				theTargetBallCenter.x,
				//	omit the y component, which should be zero
				theTargetBallCenter.z,
				theTargetBallCenter.w)
		
		return theTargetBallCenterXZW

	} else {
	
		return nil
	}
}

private func directShots(
	in gameState: NEBGameState,
	cueBallPose cueBallPoseAtStart: matrix_double3x3,
	targetBallCenters: [VectorXZW],
	allBallCenters: [VectorXZW]	//	all balls in play except the cue ball
) -> [ShotCandidate] {

	let theSpaceGeometry = gameState.spaceGeometry
	
	let theCueBallCenterAtStart = cueBallPoseAtStart * theSpaceGeometry.northPoleXZW

	//	theBallOffset geodesically translates the north pole
	//	to a point one ball shadow diameter away in the Z direction in XZW space.
	let theBallOffset = axisAlignedTranslation3D(
							axis: 1,	//	the Z axis in XZW space
							distanceInRadians: 2.0 * theSpaceGeometry.ballShadowRadius
												/ theSpaceGeometry.spaceRadius,
							geometryType: theSpaceGeometry.geometryType)
	
	//	Multiplication by theHalfTurn converts a backward looking frame
	//	at the cue ball into a forward looking frame.
	let theHalfTurn = simd_double3x3(diagonal: SIMD3<Double>(-1.0, -1.0, 1.0))
	
	var theDirectShots: [ShotCandidate] = []
	
	for theTargetBallCenter in targetBallCenters {
	
		for thePocketCenter in theSpaceGeometry.pocketCentersXZW {

			//	theFrameAtTargetBall is an orthonormal frame with
			//		axis 2 (the W axis) = theTargetBallCenter / |theTargetBallCenter|
			//		axis 1 (the Z axis) pointing away from the pocket center
			//					=> the negative Z axis points towards the pocket center
			//		axis 0 (the X axis) chosen to form a right-handed basis
			let theFrameAtTargetBall = orthonormalFrame(
										at: theTargetBallCenter,
										lookingToward: thePocketCenter,
										geometryType: theSpaceGeometry.geometryType)

			//	If anything obstructs the second leg, give up
			//	on this candidate and move on to the next one.
			if !ballPathIsClear(
					from: theFrameAtTargetBall,
					to: thePocketCenter,
					avoiding: allBallCenters,
					spaceGeometry: theSpaceGeometry) {
					
				continue
			}
			
			//	The end of the first leg should be colinear with the second leg,
			//	but one ball diameter before the second leg's beginning.
			//	(Draw a picture and all will be clear!)
			let theCueBallOutboundFrameAtImpact = theFrameAtTargetBall * theBallOffset
													//	right-to-left matrix action
			
			let theCueBallCenterAtImpact
				= theCueBallOutboundFrameAtImpact * theSpaceGeometry.northPoleXZW
			
			//	At the end of the first leg the cue ball will be
			//	at theCueBallCenterAtImpact, but its orthonormal frame
			//	relative to the first leg will be different from
			//	its orthonormal frame relative to the second leg.
			let theBackwardLookingFrame = orthonormalFrame(
										at: theCueBallCenterAtImpact,
										lookingToward: theCueBallCenterAtStart,
										geometryType: theSpaceGeometry.geometryType)
			let theCueBallInboundFrameAtImpact = theBackwardLookingFrame * theHalfTurn
													//	right-to-left matrix action
			
			//	The cue ball's momentum consists of a component parallel
			//	to the direction of impact and a component perpendicular to it.
			//	Only the parallel component gets transferred from the cue ball
			//	to the target ball.  The cosine of the incidence angle
			//	tells us what fraction of the momentum gets transferred.
			let theInboundDirection  = VectorXZW(theCueBallInboundFrameAtImpact[1])		//	-Z direction
			let theOutboundDirection = VectorXZW(theCueBallOutboundFrameAtImpact[1])	//	-Z direction
			let theMomentumTransferFactor
				= theInboundDirection.x * theOutboundDirection.x
				+ theInboundDirection.z * theOutboundDirection.z
				+ theSpaceGeometry.geometryType.geometrySign
				  * theInboundDirection.w * theOutboundDirection.w
			
			//	The dot product must be positive for the collision to be valid.
			//	If the dot product is negative, it means the cue ball would need
			//	to pass through the target ball to get to the desired point of impact
			//	and would be moving away from the target ball when it got there.
			if theMomentumTransferFactor < 0.00001 {	//	avoid nearly tangential impact
				continue
			}
			
			//	theMomentumTransferFactor may never exceed 1.0 .
			if theMomentumTransferFactor > 1.00001 {	//	allow for numerical error
				assertionFailure("Internal error: theMomentumTransferFactor > 1.0")
				continue
			}

			//	At the moment the cue stick strikes the cue ball,
			//	which way should the cue ball go?
			let theCueBallOutboundFrameAtStart = orthonormalFrame(
										at: theCueBallCenterAtStart,
										lookingToward: theCueBallCenterAtImpact,
										geometryType: theSpaceGeometry.geometryType)

			//	If anything obstructs the first leg, give up
			//	on this candidate and move on to the next one.
			if !ballPathIsClear(
					from: theCueBallOutboundFrameAtStart,
					to: theCueBallCenterAtImpact,
					avoiding: allBallCenters,
					spaceGeometry: theSpaceGeometry) {
					
				continue
			}
			
			//	theCueBallOutboundFrameAtStart and cueBallPoseAtStart
			//	share the same W axis, but differ by a rotation in the XZ plane.
			let theRotation = cueBallPoseAtStart.inverse	//	right-to-left matrix action
							* theCueBallOutboundFrameAtStart
			let theStartAngle = atan2(theRotation[0][1], theRotation[0][0])

			//	How fast would we like the target ball
			//	to be rolling when it reaches the pocket?
			let theFinalTargetBallSpeed = 0.4	//	meters/second

			//	Given theFinalTargetBallSpeed, deduce theInitialTargetBallSpeed,
			//	that is, the target ball's speed just after the cue ball hits it.
			//
			//	The function slowTheBalls() in NEBGameState.swift slows
			//	the target ball with a constant ballDeceleration,
			//	so the target ball's velocity function is
			//
			//		v(t) =  v(0)  -  a * t
			//
			//	where a = ballDeceleration (which is a positive number).
			//
			//	If we take t = 0 to be the time the ball reaches the pocket
			//	(which admittedly is an odd definition of t = 0, but makes sense
			//	in the present case because that's the time at which
			//	theFinalTargetBallSpeed is known), then the velocity function becomes
			//
			//		v(t) = fs - a * t
			//
			//	where fs = theFinalTargetBallSpeed.
			//
			//	Integrating the velocity function gives the position function
			//
			//		x(t) = x(0) + fs*t - (a/2)*t²
			//
			//	If we measure distances relative to the end of the ball's path,
			//	the position function simplifies to
			//
			//		x(t) = fs*t - (a/2)*t²
			//
			//	We can now deduce the (negative) time t at which the ball
			//	began its trip, and compute its velocity at that point.
			//	Let d be the length of the ball's path, we get
			//
			//		-d = fs*t - (a/2)*t²
			//
			//		(a/2)*t² - fs*t - d = 0
			//
			//		t = ( fs ± √( fs² + 2*a*d ) ) / a
			//
			//	We take the negative part of the '±' option;
			//	the positive part corresponds to the solution we'd get
			//	if the ball kept "accelerating backwards" and retraced its path.
			
			let theSecondLegLength = theSpaceGeometry.tableDistanceFunction(
														theTargetBallCenter,
														thePocketCenter)
									
			//	Use the quadratic formula to solve for t.
			let α₁ = 0.5 * ballDeceleration
			let β₁ = -theFinalTargetBallSpeed
			let γ₁ = -theSecondLegLength
			let t₁ = ( -β₁ - sqrt(β₁*β₁ - 4.0*α₁*γ₁) ) / (2.0 * α₁)
			
			//	What is the target ball's speed at time t + ε,
			//	immediately after the cue ball hit it?
			let theInitialTargetBallSpeed = theFinalTargetBallSpeed - ballDeceleration * t₁
			
			//	What is the cue ball's speed at time t - ε,
			//	immediately before it hit the target ball?
			let theFinalCueBallSpeed	//	meters/second
				= theInitialTargetBallSpeed / theMomentumTransferFactor

			//	The same method we used to compute theInitialTargetBallSpeed
			//	works equally well to compute theInitialCueBallSpeed
			
			let theFirstLegLength = theSpaceGeometry.tableDistanceFunction(
										theCueBallCenterAtStart,
										theCueBallCenterAtImpact)

			let α₀ = 0.5 * ballDeceleration
			let β₀ = -theFinalCueBallSpeed
			let γ₀ = -theFirstLegLength
			let t₀ = ( -β₀ - sqrt(β₀*β₀ - 4.0*α₀*γ₀) ) / (2.0 * α₀)

			let theInitialCueBallSpeed	//	meters/second
					= theFinalCueBallSpeed - ballDeceleration * t₀
			
			//	Reject a super-fast shot that would just barely graze
			//	an object ball to send it towards a pocket.
			if theInitialCueBallSpeed > 3.0 {
				continue
			}

			//	We've found a feasible shot.
			let theShotCandidate = ShotCandidate(
									angle: theStartAngle,
									startSpeed: theInitialCueBallSpeed)
			theDirectShots.append(theShotCandidate)
		}
	}

	return theDirectShots
}

private func axisAlignedTranslation3D(	//	for use in xzw coordinates (with y = 0)
	axis: Int,						//	= 0, 1 for translation in x or z direction.
	distanceInRadians θ: Double,	//	In radians, even in the Euclidean case.
	geometryType: GeometryType
) -> simd_double3x3 {	//	dimensionless, even in the Euclidean case

	precondition(
		axis == 0 || axis == 1,
		"axisAlignedTranslation3D() received an invalid axis")
	
	//	Initialize theTranslation to the identity matrix.
	var theTranslation = matrix_identity_double3x3
	
	//	Rotate the w axis ( = axis 2 ) towards the given axis.
	switch geometryType {
	
	case .spherical:
		theTranslation[axis][axis] =  cos(θ)
		theTranslation[axis][ 2  ] = -sin(θ)
		theTranslation[ 2  ][axis] =  sin(θ)
		theTranslation[ 2  ][ 2  ] =  cos(θ)
		
	case .euclidean:
		//	Just as in the spherical and hyperbolic cases,
		//	in the Euclidean case too
		//
		//		θ = (the translation distance) / (the space radius)
		//
		//	A product like
		//
		//		       (  1     0    0 )
		//		(x z ρ)(  0     1    0 )
		//		       ( Δx/ρ  Δz/ρ  1 )
		//
		//		where ρ = (the space radius),
		//
		//	gives the expected result.
		//
		theTranslation[axis][axis] = 1.0
		theTranslation[axis][ 2  ] = 0.0
		theTranslation[ 2  ][axis] =  θ
		theTranslation[ 2  ][ 2  ] = 1.0
		
	case .hyperbolic:
		theTranslation[axis][axis] = cosh(θ)
		theTranslation[axis][ 2  ] = sinh(θ)
		theTranslation[ 2  ][axis] = sinh(θ)
		theTranslation[ 2  ][ 2  ] = cosh(θ)
	}
	
	return theTranslation
}

private func orthonormalFrame(
	at pointA: VectorXZW,				//	in meters in XZW space
	lookingToward pointB: VectorXZW,	//	in meters in XZW space
	geometryType: GeometryType
) -> simd_double3x3
	//	Returns a dimensionless orthonormal frame with
	//		axis 2 (the W axis) = pointA / |pointA|
	//		axis 1 (the Z axis) pointing away from pointB
	//					=> the negative Z axis points towards pointB
	//		axis 0 (the X axis) chosen to form a right-handed basis
{

	let verticalMetric   = geometryType.xzwMetricForVertical
	let horizontalMetric = geometryType.xzwMetricForHorizontal

	let wAxis = verticalMetric.normalize(pointA)
	
	let theRawDifference = pointA - pointB
	let theProjectionLength = verticalMetric.innerProduct(
												theRawDifference,
												wAxis)
	let theProjection = theProjectionLength * wAxis
	let theUnnormalizedZAxis = theRawDifference - theProjection
	let zAxis = horizontalMetric.normalize(theUnnormalizedZAxis)
	
	//	In all three geometries (spherical, Euclidean, hyperbolic)
	//	the following formula for the cross product yields a vector that
	//
	//		is orthogonal to both zAxis and wAxis,
	//
	//		has dimensionless length 1
	//			(if the two input vectors weren't orthonormal,
	//			a more general formula would apply), and
	//
	//		gives a basis (xAxis, zAxis, wAxis) with
	//			the same handedness as the XZW coordinate system.
	//
	//	Note:  I've checked that those claims are correct
	//	in the present case, but it would need modifications
	//	in other cases, for example, the case of two "horizontal"
	//	vectors in Euclidean geometry would require the use
	//	of the "vertical" Euclidean metric.
	//
	let σ = geometryType.geometrySign
	let xAxis = VectorXZW(
			zAxis.z * wAxis.w  -  zAxis.w * wAxis.z,
			zAxis.w * wAxis.x  -  zAxis.x * wAxis.w,
	   σ * (zAxis.x * wAxis.z  -  zAxis.z * wAxis.x) )

	let theOrthonormalFrame = simd_double3x3(
								xAxis.v,
								zAxis.v,
								wAxis.v)

	return theOrthonormalFrame
}

private func ballPathIsClear(
	from ballFrameAtStart: simd_double3x3,
	to ballCenterAtFinish: VectorXZW,
	avoiding ballsToAvoid: [VectorXZW],
	spaceGeometry: NEBSpaceGeometry
) -> Bool {

	let theWorldToFrame = ballFrameAtStart.inverse

	let theBallCenterAtFinishInFrameSpace = theWorldToFrame * ballCenterAtFinish

	for ballToAvoid in ballsToAvoid {
	
		let theAvoidedBallInFrameSpace = theWorldToFrame * ballToAvoid
	
		if !pathFromNorthPole(
			to: theBallCenterAtFinishInFrameSpace,
			avoidsBallAt: theAvoidedBallInFrameSpace,
			spaceGeometry: spaceGeometry) {
		
			return false
		}
	}
	
	return true
}

//	pathFromNorthPole() works entirely in the ball's starting frame.
//	So the ball's path begins at the north pole and runs
//	in the negative Z direction to the ballCenterAtFinish.
//	The otherBallCenter is, of course, given in this same
//	coordinate system.
private func pathFromNorthPole(
	to ballCenterAtFinish: VectorXZW,
	avoidsBallAt otherBallCenter: VectorXZW,
	spaceGeometry: NEBSpaceGeometry
) -> Bool {

	let ρ = spaceGeometry.spaceRadius
	
	//	Please see the extended comment in ballShadowRadius()
	//	in NEBSpaceGeometry.swift for the definition of a ball's
	//	orthogonal shadow.
	let r = spaceGeometry.ballShadowRadius

	//	As the ball moves along a geodesic from the start to the finish,
	//	its orthogonal shadow sweeps out a capsule-like shape,
	//	with circular endcaps of radius r and sides that are
	//	not quite geodesics but instead are curves running
	//	a constant distance r from the path itself.
	//	We want to check whether the ballToAvoid's orthogonal shadow
	//	overlaps that capsule.  This is equivalent to asking whether
	//	the otherBallCenter sits within a capsule with endcap radius 2r
	//	and width 2r.
	
	//	We're working in the ball's starting frame, so the path
	//	begins at (x,z,w) = (0,0,ρ) and runs in the direction
	//	of the negative Z axis.  If the otherBallCenter lies
	//	further than 2r meters away from that path, we know that
	//	the other ball's orthogonal shadow can't overlap the capsule
	//	described above.  Because the path runs in the negative Z direction,
	//	the distance to that path depends only on the otherBallCenter's
	//	x coordinate.
	
	let xMin: Double
	switch spaceGeometry.geometryType {
	
	case .spherical:
		xMin = ρ * sin( 2.0 * r / ρ )
		
	case .euclidean:
		xMin = 2.0 * r	//	= ρ * ( 2.0 * r / ρ )
		
	case .hyperbolic:
		xMin = ρ * sinh( 2.0 * r / ρ )
	}
	
	if abs(otherBallCenter.x) > xMin {
		return true
	}
	
	//	We now know that the otherBallCenter lies somewhere
	//	in the ±2r band running along the path, but it might
	//	lie before the path's start or after its finish.
	
	//	The only ball whose center can fall within the 2r-radius disk
	//	centered at the path's start is the ball whose path
	//	we're following -- we expect it to be there!
	let theDistanceToStartPoint
		= spaceGeometry.tableDistanceFunction(
							otherBallCenter,
							spaceGeometry.northPoleXZW)
	if theDistanceToStartPoint < 2.0 * r - 0.001 {
		assert(
			theDistanceToStartPoint < 0.001,
			"Other ball partially overlaps starting ball, or else numerical precision is terrible.")
		return true	// other ball = this ball
	}
	
	//	Test whether the other ball's center falls within
	//	the 2r-radius disk at the end of the path.
	//	Allow a little bit of slack, because when it's
	//	the cue ball following the path, we expect it to be
	//	exactly tangent to a target ball at the path's end.
	
	let theDistanceToFinishPoint
		= spaceGeometry.tableDistanceFunction(
							otherBallCenter,
							ballCenterAtFinish)
	if theDistanceToFinishPoint < 2.0 * r - 0.001 {
		return false
	}
	
	//	We know from above that the other ball's orthogonal shadow
	//	overlaps the ±2r band running along the path, but it
	//	doesn't over the 2r-radius disks at either end of the capsule.
	//	So to test whether it overlaps the capsule at all,
	//	it suffices to test whether its "distance along the path"
	//	falls between the path's start and finish.
	//
	//	tn(θ) represents
	//	   tan(θ) in the spherical case
	//	     θ    in the Euclidean case
	//	  tanh(θ) in the hyperbolic case
	//
	let tnθ = otherBallCenter.z / otherBallCenter.w
	let maxTnθ = 0.0
	let minTnθ = ballCenterAtFinish.z / ballCenterAtFinish.w	//	< 0.0
	assert(
		minTnθ < 0.0,
		"Internal error:  minTnθ is non-negative")
	if tnθ > minTnθ && tnθ < maxTnθ {
		return false
	}

	return true
}

private func chooseHailMaryShot(
	in gameState: NEBGameState,
	cueBallPose: matrix_double3x3?,
	targetBallCenters: [VectorXZW]
) -> ShotCandidate {

	//	Shoot towards the nearest target ball, with no regard
	//	for whether the path is clear nor for what might happen
	//	if the cue ball does indeed reach the closest ball.
	
	let theSpaceGeometry = gameState.spaceGeometry
	
	guard let theCueBallPose = cueBallPose else {
		assertionFailure("chooseHailMaryShot() called with cueBallPose = nil")
		return ShotCandidate(angle: 0.0, startSpeed: 0.01)
	}
	let theCueBallCenter = theCueBallPose * theSpaceGeometry.northPoleXZW
	
	func distanceToTargetBallCenter(
		_ targetBallCenter: VectorXZW
	) -> Double {
	
		let theDistance = theSpaceGeometry.tableDistanceFunction(
											theCueBallCenter,
											targetBallCenter)
		return theDistance
	}
	
	guard let theFirstTargetBallCenter = targetBallCenters.first else {
		assertionFailure("chooseHailMaryShot() called with targetBallCenters empty")
		return ShotCandidate(angle: 0.0, startSpeed: 0.01)
	}
	var theClosestBallCenter = theFirstTargetBallCenter
	var theSmallestDistance = distanceToTargetBallCenter(theClosestBallCenter)
	
	for theTargetBallCenter in targetBallCenters {
	
		let theDistance = distanceToTargetBallCenter(theTargetBallCenter)
		
		if theDistance < theSmallestDistance {
			theClosestBallCenter = theTargetBallCenter
			theSmallestDistance = theDistance
		}
	}

	//	Shoot the cue ball towards theClosestBallCenter,
	//	with no regard for what happens next.
	//	Just shoot and hope for the best.
	
	let theCueBallOutboundFrame = orthonormalFrame(
									at: theCueBallCenter,
									lookingToward: theClosestBallCenter,
									geometryType: theSpaceGeometry.geometryType)
	let theRotation = theCueBallPose.inverse	//	right-to-left matrix action
					* theCueBallOutboundFrame
	let theAngle = atan2(theRotation[0][1], theRotation[0][0])

	return ShotCandidate(angle: theAngle, startSpeed: 2.0)
}

private func slowestShot(
	of shotCandidates: [ShotCandidate]	//	shotCandidates must not be empty
) -> ShotCandidate {
	
	//	Select an arbitrary ShotCandidate to initialize the reduce() call below.
	guard let theFirstShotCandidate = shotCandidates.first else {
		preconditionFailure("Internal error: shotCandidates() is unexpectedly empty")
	}
	
	let theSlowestShot = shotCandidates.reduce(theFirstShotCandidate) {
		theBestSoFar, theCandidate in
		
		return theCandidate.startSpeed < theBestSoFar.startSpeed ?
				theCandidate : theBestSoFar
	}
	
	return theSlowestShot
}

private func applyGaussianError(
	to preciseShot: ShotCandidate,
	difficulty: Double	//	∈ 0.0 (easy) ... 1.0 (hard)
) -> ShotCandidate {

	//	When the difficulty is "easy",
	//		we introduce the largest errors into the computer's shots.
	//	When the difficulty is "hard"
	//		we introduce no errors at all.
	let theEasiness = 1.0 - difficulty	//	∈ 0.0 (hard) ... 1.0 (easy)

	let theAngleStdDev    = theEasiness * maxAngleStdDev	//	radians
	let theLogSpeedStdDev = theEasiness * maxLogSpeedStdDev	//	dimensionless

	let theAngleError    =    theAngleStdDev * getRandomGaussian()
	let theLogSpeedError = theLogSpeedStdDev * getRandomGaussian()

	let theImpreciseAngle = preciseShot.angle + theAngleError
	let theImpreciseSpeed = exp( log(preciseShot.startSpeed) + theLogSpeedError )

	let theImpreciseShot = ShotCandidate(
							angle: theImpreciseAngle,
							startSpeed: theImpreciseSpeed)

	return theImpreciseShot
}


// MARK: -
// MARK: Place cue ball near center

func validCueBallPlacementNearCenter(
	in gameState: NEBGameState
) -> simd_double4x4 {
	
	//	Someday we might want to place the cue ball in some
	//	strategically chosen spot, for example one from which
	//	it would be easy to sink one of our assigned target balls.
	//	For now, though, let's just try a bunch of arbitary places
	//	until we find one that's unobstructed by other balls.

	let theSpaceGeometry = gameState.spaceGeometry
	
	let theBallCenters = getAllBallCentersXZW(in: gameState)	//	excludes cue ball

	if cueBall(
			at: theSpaceGeometry.northPoleXZW,
			avoids: theBallCenters,
			spaceGeometry: theSpaceGeometry) {
			
		return matrix_identity_double4x4
	}
	
	//	Try some moderately spaced positions.
	//	There's nothing scientific about numRadialSteps or numAzimuthalSteps
	//	-- it's all pretty arbitrary.
	//
	let numRadialSteps = 5
	for i in 1...numRadialSteps {	//	number of radial steps
	
		let r = Double(i) * theSpaceGeometry.ballShadowRadius
		
		let theCircumference = NEBSpaceGeometry.computeCircleCircumference(
												radius: r,
												geometryType: theSpaceGeometry.geometryType,
												spaceRadius: theSpaceGeometry.spaceRadius)
		let numAzimuthalSteps = floor( theCircumference / theSpaceGeometry.ballShadowRadius )
		let Δθ = 2.0 * Double.pi / numAzimuthalSteps
		
		for j in 0 ..< Int(numAzimuthalSteps) {
		
			let θ = Double(j) * Δθ
			
			let theProtoTranslation = r * SIMD3<Double>(cos(θ), 0.0, sin(θ))
			
			let theCueBallPlacement = geodesicTranslation(
				v: theProtoTranslation / theSpaceGeometry.spaceRadius,
				geometryType: theSpaceGeometry.geometryType)
			
			let theCueBallCenter = theCueBallPlacement * theSpaceGeometry.northPole
			let theCueBallCenterXZW = VectorXZW(
				theCueBallCenter.x,	//	x
				//	omit y
				theCueBallCenter.z,	//	z
				theCueBallCenter.w )	//	w
			
			if cueBall(
				at: theCueBallCenterXZW,
				avoids: theBallCenters,
				spaceGeometry: theSpaceGeometry) {
				
				return theCueBallPlacement
			}
		}
	}
	
	preconditionFailure("Unexpectedly failed to find a valid place to put the cue ball in chooseCueBallPlacement()")
}

func cueBall(
	at cueBallCenterXZW: VectorXZW,
	avoids ballCentersXZW: [VectorXZW],
	spaceGeometry: NEBSpaceGeometry
) -> Bool {	//	Does the candidate cue ball position avoid the other balls?

	let theMinDistance = 2.0 * spaceGeometry.ballShadowRadius

	for theBallCenterXZW in ballCentersXZW {
	
		if spaceGeometry.tableDistanceFunction(
							cueBallCenterXZW,
							theBallCenterXZW)
			<= theMinDistance {
		
			return false	//	cue ball would overlap this other ball
		}
	}
	
	return true	//	cue ball avoids all other balls
}

func getRandomGaussian() -> Double {	//	Returns values sampled from a Gaussian distribution
										//		with mean 0.0 and standard deviation 1.0

	//	Use the Box-Muller transform to generate
	//	a normally distributed random variable.

	var u₁: Double
	repeat {
		u₁ = Double.random(in: 0.0 ..< 1.0)	//	uniform in 0.0 ≤ u₁ ≤ 1.0
	} while u₁ == 0.0	//	almost never occurs
	
	let u₂ = Double.random(in: 0.0 ..< 1.0)	//	uniform in 0.0 ≤ u₂ ≤ 1.0

	//	x values will be Gaussian with mean 0.0 and standard deviation 1.0
	let x = sqrt( -2.0 * log(u₁) ) * cos( 2.0 * Double.pi * u₂ )

	return x
}
