//	NEBControlPanel.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


//	I would have liked to allow a broader range of table sizes here,
//	so that players could experience curvature on a variety of scales,
//	but we can't let the table get too large without running afoul
//	of visionOS's safety feature that fades players back to real reality
//	whenever they wander more than 1.5 m from their starting point.
let minTableEdgeLengthInMeters		= 0.5
let maxTableEdgeLengthInMeters		= 1.5	//	corner-to-center = (1.5/2)√2 = ~1.06 m
let defaultTableEdgeLengthInMeters	= 1.0	//	same even with forTalks

let minTableHeight		= 0.7	//	in meters, above the floor

//	When there's an elevated stage, visionOS can get confused
//	about where the floor is. When this happened in Leipzig in 2024,
//	the pool table ended up ~20 cm lower than it was supposed to be,
//	forcing players to bend over to play. In case this ever happens
//	again anywhere, let's allow the table to be much higher than
//	would be reasonable for normal home use.
let maxTableHeight		= forTalks ? 1.5 : 1.0	//	in meters, above the floor

#if targetEnvironment(simulator)

//	The only reason for supporting the simulator is that
//	it's a convenient environment for making screenshots
//	of the Control Panel. You can't actually play pool
//	in the simulator because it doesn't support hand tracking.

//	The simulator seems to be placing the coordinate origin at the player's eyes,
//	in contrast to the Vision Pro itself which (as expected and as documented)
//	places the coordinate origin at the player's feet.
let defaultTableHeight	= -0.7	//	in meters, "above" the player's eyes, i.e. below them

#else

//	A regulation billiards table must have a bed that sits
//	between 29.25" and 31" above the floor (roughly 74-79 cm).
//	To use it, players typically extend their bodies
//	horizontally and rest some weight on the table.
//	But Curved-Space Pool Hall uses no physical table,
//	and has no place for the player to rest any weight.
//	So let's assume the player will be standing more-or-less
//	vertical and will prefer a slightly higher table.
//
//	A higher table is also more convenient in hyperbolic space,
//	where the player perceives the table as sloping downwards
//	(as it goes away from the player).
//
//	In addition, during lectures a regulation-height table
//	is too low to be viewed comfortably edge-on.
//	So let's default to a 80 cm default table height.
//
let defaultTableHeight	= forTalks ? 0.90 : 0.80
							//	in meters, above the floor
							//	(use 0.76 for regulation height)

#endif


struct NEBControlPanel: View {

	@Binding var playerHandedness: PlayerHandedness
	@Binding var computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
	@Binding var guardianAngelFactor: Double		//	∈ 0.0 ... 1.0
	@Binding var geometryType: GeometryType
	@Binding var tableEdgeLengthInMeters: Double
	@Binding var tableHeight: Double
	@Binding var adjustPerceivedDistances: Bool
	@Binding var showHints: Bool
	@Binding var instructionsViewIsPresented: Bool
	@Binding var instructionItem: InstructionItem
	@Binding var hideHands: Bool

	@State var showSettingsView: Bool = false

	@Environment(\.dismissWindow) var dismissWindow
	@Environment(\.openImmersiveSpace) var openImmersiveSpace

	var body: some View {
	
		//	The order {euc, hyp, sph} encourages players to try
		//	the three geometries in ascending order of difficulty,
		//	which is probably more important than presenting them
		//	in their geometrically natural order {sph, euc, hyp}.
		//
		let theRecommendedGeometryOrder: [GeometryType] = [.euclidean, .hyperbolic, .spherical]

		let dismissControlPanel = {
			dismissWindow(id: "Control panel")
			gInstructionsPanelHasClosed.set(true)
		}
	
		VStack(spacing: 24.0) {

			VStack(spacing: 12.0) {
			
				ForEach(theRecommendedGeometryOrder, id: \.id) { theGeometryType in
				
					Button() {
					
						//	Set the @State variable for the requested GeometryType.
						geometryType = theGeometryType

						Task {

							//	The comment accompanying hideHands in NEBApp
							//	explains why we need to open the immersive space
							//	with the player's hands visible and then wait
							//	for it to appear before hiding the hands.
							//
							hideHands = false
							
							//	Suppress the cue stick until the player
							//	finishes reading the instructions panels.
							gInstructionsPanelHasClosed.set(false)
						
							if await openImmersiveSpace(id: "Billiards immersive space") == .opened {

								instructionItem = .walkBackwards
								instructionsViewIsPresented = true

								//	If we set hideHands=true too soon, our request
								//	sometimes gets ignored. So let's wait a second.
								//
								//	As an added bonus, it's good for the player
								//	to see his/her physical hands briefly overlapping
								//	the virtual hands before the former fade away.
								//
								try await Task.sleep(nanoseconds: 1_000_000_000)
								
								hideHands = true

							} else {
								assertionFailure("Failed to create billiards immersive space")
							}
						}
						
					} label: {
					
						if forTalks {
						
							ZStack() {
							
								let theColor = theGeometryType.playButtonDisplayP3ColorForTalks
							
								Circle()
								.frame(width: 256.0, height: 256.0)
								.foregroundColor(Color(
													.displayP3,
													red: theColor.x,
													green: theColor.y,
													blue: theColor.z,
													opacity: 1.0))
								
								Text(theGeometryType.playButtonTitleForTalks)
									.font(.system(size: 96.0))
								.foregroundColor(.black)
							}
							.frame(maxWidth: .infinity)
							
						} else {	//	standard button
						
							Text(theGeometryType.playButtonTitle)
							.multilineTextAlignment(.center)
							.frame(maxWidth: .infinity)
						}
					}
				}
			}
			.sheet(isPresented: $instructionsViewIsPresented,
				onDismiss: {
					Task { dismissControlPanel() }
				},
				content: {
					InstructionsView(
						playerHandedness: playerHandedness,
						instructionItem: $instructionItem,
						showHints: $showHints)
				}
			)
			
			if forTalks {
				Button() {
					playerHandedness.toggle()
				} label: {
					ZStack() {
					
						Circle()
						.frame(width: 256.0, height: 256.0)
						.foregroundColor(Color(white: 0.875))
						
						HStack() {
							Text("L")
							.foregroundColor(playerHandedness == .leftHanded ?
													.black : Color(white: 0.75))
							Text("/")
							.foregroundColor(Color(white: 0.75))
							Text("R")
							.foregroundColor(playerHandedness == .rightHanded ?
													.black : Color(white: 0.75))
						}
						.font(.system(size: 96.0))
						.foregroundColor(.black)
					}
					.frame(maxWidth: .infinity)
				}
			}

			HStack() {

				Button(
					action: { showSettingsView = true },
					label: { Image(systemName: "gear") }
				)
				.sheet(isPresented: $showSettingsView) {
					SettingsView(
						playerHandedness: $playerHandedness,
						computerPlayerDifficulty: $computerPlayerDifficulty,
						guardianAngelFactor: $guardianAngelFactor,
						tableEdgeLengthInMeters: $tableEdgeLengthInMeters,
						tableHeight: $tableHeight,
						adjustPerceivedDistances: $adjustPerceivedDistances,
						showHints: $showHints)
				}
			}
			.frame(maxWidth: .infinity, alignment: .center)
		}
		.padding(32.0)
	}
}

struct SettingsView: View {

	@Binding var playerHandedness: PlayerHandedness
	@Binding var computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
	@Binding var guardianAngelFactor: Double		//	∈ 0.0 ... 1.0
	@Binding var tableEdgeLengthInMeters: Double
	@Binding var tableHeight: Double
	@Binding var adjustPerceivedDistances: Bool
	@Binding var showHints: Bool

	@State var showComputerOppenentInfo: Bool = false
	@State var showGuardianAngelInfo: Bool = false
	@State var showPerceivedDistancesInfo: Bool = false

	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	var body: some View {

		Grid(horizontalSpacing: 24.0, verticalSpacing: 24.0) {
		
			GridRow() {
				HStack() {
					Text("Player is")
					Spacer()
					Picker("Player handedness", selection: $playerHandedness) {
						ForEach(PlayerHandedness.allCases, id: \.id) { thePlayerHandedness in
							Text(thePlayerHandedness.name).tag(thePlayerHandedness)
						}
					}
				}
			}

			GridRow() {

				VStack(alignment: .leading) {

					HStack() {
					
						Text("Computer opponent")
						
						Spacer()
						
						Text(
							difficultyDescription()
						  + String(format:" (%d %%)", Int(computerPlayerDifficulty * 100.0))
										//	uses 'thin space'
						)
						.foregroundColor(.secondary)
					}
				
					Slider(
						value: $computerPlayerDifficulty,
						in: 0.0 ... 1.0)
				}
					
				Button() {
					showComputerOppenentInfo = true
				} label: {
					Image(systemName: "info.circle")
				}
				.sheet(isPresented: $showComputerOppenentInfo) {
					ComputerOppenentInfo()
				}
			}

			GridRow() {
			
				VStack() {
								
					HStack() {
					
						Text("Guardian angel")
						
						Spacer()
						
						Text(String(format:"%d %%", Int(guardianAngelFactor * 100.0)))
									//	uses 'thin space'
						.foregroundColor(.secondary)
					}
				
					Slider(
						value: $guardianAngelFactor,
						in: 0.0 ... 1.0)
				}
					
				Button() {
					showGuardianAngelInfo = true
				} label: {
					Image(systemName: "info.circle")
				}
				.sheet(isPresented: $showGuardianAngelInfo) {
					GuardianAngelInfo()
				}
			}

			GridRow() {
				VStack() {
					HStack() {
					
						Text("Table edge length")
						
						Spacer()
						
						Text(String(format:"%.1f cm", tableEdgeLengthInMeters * 100.0))
						.foregroundColor(.secondary)
					}
					Slider(
						value: $tableEdgeLengthInMeters,
						in: minTableEdgeLengthInMeters ... maxTableEdgeLengthInMeters)
				}
			}
		
			GridRow() {
				VStack() {
					HStack() {
					
						Text("Table height")
						
						Spacer()
						
						Text(String(format:"%.1f cm", tableHeight * 100.0))
						.foregroundColor(.secondary)
					}
					Slider(
						value: $tableHeight,
						in: minTableHeight ... maxTableHeight)
				}
			}

			GridRow() {
			
				Toggle(isOn: $adjustPerceivedDistances) {
					Text("Adjust perceived distances")
				}
				
				Button() {
					showPerceivedDistancesInfo = true
				} label: {
					Image(systemName: "info.circle")
				}
				.sheet(isPresented: $showPerceivedDistancesInfo) {
					PerceivedDistancesInfo()
				}
			}

			GridRow() {
				Toggle(isOn: $showHints) {
					Text("Show hints when starting game")
				}
			}

			Button("Done") {
				dismiss()
			}
		}
		.monospacedDigit()
		.pickerStyle(.segmented)
		.padding(EdgeInsets(top: 32.0, leading: 64.0, bottom: 32.0, trailing: 32.0))
	}

	func difficultyDescription() -> String {
	
		if computerPlayerDifficulty == 0.0 {
			return "Weak"
		} else
		if computerPlayerDifficulty == 1.0 {
			return "Strong"
		} else {
			return "Intermediate"
		}
	}
}

struct InstructionsView: View {

	         let playerHandedness: PlayerHandedness
	@Binding var instructionItem: InstructionItem
	@Binding var showHints: Bool

	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	@State var translatePerson: Bool = false
	
	let rotationData = [
		(-0.1, "Spherical",  Color(.displayP3, red: 0.25, green: 0.00, blue: 1.00)),
		( 0.0, "Euclidean",  Color(.displayP3, red: 0.00, green: 0.75, blue: 1.00)),
		(+0.1, "Hyperbolic", Color(.displayP3, red: 0.00, green: 0.75, blue: 0.50))
	]

	var body: some View {
	
		VStack(spacing: 32.0) {
		
			switch instructionItem {
			
			case .walkBackwards:

				HStack(spacing: 16.0) {

					ZStack() {

						Image("pool table")
						.resizable()
						.scaledToFit()
						.frame(width: 72.0, height: 72.0)
						.frame(width: 96.0, height: 96.0, alignment: .top)

						Image("person")
						.resizable()
						.scaledToFit()
						.frame(width: 48.0, height: 48.0)
						.offset(y: translatePerson ? +24.0 : -24.0)
						.frame(width: 96.0, height: 96.0)
						.task {
							withAnimation( .linear(duration: 2.0).delay(1.0) ) {
								translatePerson = true
							}
						}
					}
					
					Text("Please walk backwards until\nyou see the whole table in front of you.")
					.fixedSize(horizontal: false, vertical: true)
						//	fixedSize() is needed to allow multiple lines
						//	if the enclosing frame's height isn't specified
						//	(unintuitive but true)
					.multilineTextAlignment(.center)
				}

				DisclosureGroup(
					content: {
						Text("visionOS limits you to a radius of 1.5 meters from your starting point — if you walk any further it will fade you back to real reality, for safety reasons. Curved-Space Pool Hall centers the pool table within that 1.5 m safe circle, so you can walk freely all the way around the table without leaving the simulation. This works fine during play, but it means that each time you start a new game you find yourself at the center of the safe circle — and therefore at the center of the table — which is why you need to take a few steps backwards before you take your first shot.")
						.font(.caption2)
						.fixedSize(horizontal: false, vertical: true)
							//	fixedSize() is needed to allow multiple lines
							//	if the enclosing frame's height isn't specified
							//	(unintuitive but true)
						.multilineTextAlignment(.leading)
						.padding(16.0)
					},
					label: {
						Text("Why did the app start me in the middle of the table?")
						.font(.caption2)
					}
				)
				.frame(width: 416.0)
			
			case .handsAboveTable:
			
				Image("cue ball in hand")
				.resizable()
				.scaledToFit()
				.frame(width: 148.0, height: 108.0)
				.scaleEffect(x: playerHandedness == .leftHanded ? +1.0 : -1.0)

				Text("Raise your \(playerHandedness == .leftHanded ? "right" : "left") hand to get a cue ball.")
				.fixedSize(horizontal: false, vertical: true)
					//	fixedSize() is needed to allow multiple lines
					//	if the enclosing frame's height isn't specified
					//	(unintuitive but true)
				.multilineTextAlignment(.center)
			
			case .placeCueBall:
			
				ZStack() {

					Image("pool table")
					.resizable()
					.scaledToFit()
					.frame(width: 160.0, height: 160.0, alignment: .bottom)

					Image(systemName: "circle.fill")
					.resizable()
					.scaledToFit()
					.frame(width: 24.0, height: 24.0)
					.offset(x: -32.0, y: +16.0)
					.frame(width: 160.0, height: 160.0)
				}

				Text("You can place the cue ball\non the table anywhere you want.")
				.fixedSize(horizontal: false, vertical: true)
					//	fixedSize() is needed to allow multiple lines
					//	if the enclosing frame's height isn't specified
					//	(unintuitive but true)
				.multilineTextAlignment(.center)

			case .gripHandHeight:

				VStack(spacing: 3.0) {
				
					Text("Hint")
					.font(.title2)
					.fixedSize(horizontal: false, vertical: true)
						//	fixedSize() is needed to allow multiple lines
						//	if the enclosing frame's height isn't specified
						//	(unintuitive but true)
					.multilineTextAlignment(.center)

					Text("Just before you shoot, take a quick glance\nat the back end of the cue stick. Make sure\nthe whole stick lies parallel to the table.")
					.fixedSize(horizontal: false, vertical: true)
						//	fixedSize() is needed to allow multiple lines
						//	if the enclosing frame's height isn't specified
						//	(unintuitive but true)
					.multilineTextAlignment(.center)
				}

				ZStack() {
					ForEach(0...2, id: \.self) { i in
					
						ZStack(alignment: .trailing) {
						
							Image("cue stick")
							.resizable()
							.scaledToFit()
							.frame(width: 512.0, height: 16.0)
							
							Text(rotationData[i].1)
							.foregroundStyle(rotationData[i].2)
							.frame(width: 96.0)
							.padding(.trailing, 64.0)
						}
						.rotationEffect(.radians(rotationData[i].0), anchor: .leading)
					}
				}
				.padding([.top, .bottom], 32.0)

				Text("In hyperbolic pool you'll need to keep your grip hand lower than you do in Euclidean pool, and in spherical pool you'll need to keep it higher. Otherwise you may find your cue stick passing above or below the cue ball when you try to take your shot.")
				.font(.caption2)
				.fixedSize(horizontal: false, vertical: true)
					//	fixedSize() is needed to allow multiple lines
					//	if the enclosing frame's height isn't specified
					//	(unintuitive but true)
				.multilineTextAlignment(.center)

				//	Once the cue stick is parallel to the table,
				//	the player then needs to translate the cue stick
				//	along its central geodesic. But that's not
				//	an easy thing to say to non-mathematicians
				//	in a short clear sentence. So let's leave
				//	that unsaid, at least for now.

				Button("I'll do that") {
					instructionItem = .howToReturn
				}
				
				DisclosureGroup(
					content: {
						Text("Quick answer: That's the way non-Euclidean geometry works.\nTechnical answer: It's a consequence of geodesic divergence or convergence.")
						.font(.caption2)
						.fixedSize(horizontal: false, vertical: true)
							//	fixedSize() is needed to allow multiple lines
							//	if the enclosing frame's height isn't specified
							//	(unintuitive but true)
						.multilineTextAlignment(.center)
						.padding(16.0)
					},
					label: {
						Text("Why?")
						.font(.caption2)
					}
				)
			
			case .howToReturn:

				VStack(spacing: 16.0) {
				
					Image("press Digital Crown")
					.resizable()
					.frame(width: 105.0, height: 90.0)
					
					Text("When you're done playing, you can\npress the Digital Crown to end the game\nand return to the control panel.")
					.fixedSize(horizontal: false, vertical: true)
						//	fixedSize() is needed to allow multiple lines
						//	if the enclosing frame's height isn't specified
						//	(unintuitive but true)
					.multilineTextAlignment(.center)
				}
				
				Button("I'll remember that") {
					instructionItem = .startPlaying
				}
			
			case .startPlaying:
				
				Image("cue stick")
				.resizable()
				.scaledToFit()
				.frame(width: 512.0, height: 16.0)

				Text("When you start playing you'll find\na cue stick in your hands.")
				.fixedSize(horizontal: false, vertical: true)
					//	fixedSize() is needed to allow multiple lines
					//	if the enclosing frame's height isn't specified
					//	(unintuitive but true)
				.multilineTextAlignment(.center)
				
				Button("Start playing!") {
					dismiss()
				}

				Toggle(isOn: Binding<Bool>(
					get: { !showHints },
					set: { dontShowAgain in showHints = !dontShowAgain }))
				{
					Text("Don't show hints again")
					.font(.caption)
				}
				.fixedSize()
			}
		}
		.padding(32.0)
		.frame(maxWidth: 592.0)
	}
}

struct ComputerOppenentInfo: View {

	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	var body: some View {
	
		VStack(spacing: 32.0) {
			
			Text("Computer opponent")
			.font(.title)
		
			Text("Your computer opponent considers all possible direct shots (no combination shots) and chooses the one it likes best. The computer could of course easily make that shot, sinking the object ball every time. And if you set it to *Strong (100%)*, that's exactly what it will do. But that's no fun for you, the human, to just watch the computer win every time. Instead you can adjust the computer's strength for a more interesting game. As you push the slider towards *Weak (0%)*, the app adds a progressively larger “Gaussian random error” to the computer's shots, making it play more like another human.")
			
			Button("Dismiss") {
				dismiss()
			}
		}
		.padding(32.0)
		.frame(maxWidth: 512.0)
	}
}

struct GuardianAngelInfo: View {

	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	var body: some View {
	
		VStack(spacing: 32.0) {
			
			Text("Guardian angel")
			.font(.title)
		
			Text("The guardian angel tries to see which ball you're attempting to sink in which pocket. If you've set the guardian angel to 100%, it removes all the error from your shot and you sink the ball exactly. By contrast, if you've set it to 0% it lets your shot proceed unmodified — you're on your own! Otherwise, with the guardian angel set to some intermediate strength, it interpolates between your actual shot and the perfect shot that you intended.")
			
			Button("Dismiss") {
				dismiss()
			}
		}
		.padding(32.0)
		.frame(maxWidth: 512.0)
	}
}
