//	NEBPoseMapper.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

class NEBPoseMapper {

	//	Note:  The following paragraphs are self-contained.
	//	Nevertheless, for an illustrated and more complete
	//	exposition of these ideas, please see §2 and §3 of
	//
	//		Jeff Weeks,
	//		Body coherence in curved-space virtual reality games,
	//		Computers & Graphics 97 (2021) 28–41.
	//
	//		  https://www.sciencedirect.com/science/article/pii/S0097849321000443
	//		or
	//		  https://arxiv.org/abs/2011.00510
	//
	
	//	A naive way to model headset motion in a potentially curved
	//	3-dimensional space (that is, in the 3-sphere, Euclidean 3-space
	//	or hyperbolic 3-space) would be, for each update event
	//	(typically 90 times per second) to simply take the small
	//	incremental change to the headset's pose in the physical room,
	//	express it in the headset's own local coordinate system,
	//	and apply that same local change to the "virtual head pose"
	//	in the simulation space.  Alas, in the presence of holonomy,
	//	that naive approach would be problematic. For example,
	//	if you're in a simulation of hyperbolic 3-space and you
	//	walk around in a small circle, the virtual billiards table
	//	would gradually rotate around you and would eventually be
	//	behind your back, with some or all of the table lying
	//	outside visionOS's 1.5 meter "safe zone". If you then left
	//	the safe zone to reach the table's far side, visionOS would
	//	fade out the immersive space and fade in the passthrough video
	//	instead.
	//
	//	To avoid that problem, we'll make sure that the player's
	//	virtual head always maintains the "same" pose relative
	//	to an edge of the virtual table that his/her physical head
	//	maintains relative to an edge of an implicit square physical table.
	//
	//	Our algorithm will be to express the player's pose
	//	in the physical room relative to a coordinate system attached
	//	to a nearby point on the edge of the implicit square physical table,
	//	then transfer that pose to the virtual space in an analogous way,
	//	mapping it first into the tangent space attached to an appropriate
	//	point on the virtual table's edge, and thence into curved space itself..
	//
	//	What point on the table edge should we base
	//	the aforementioned tangent space at?  My first thought
	//	had been to use the closest point on the closest edge,
	//	but after thinking more carefully I realized that this would
	//	introduce a small discontinuity: if the player is near a corner,
	//	with his/her head over the table itself, then as his/her head
	//	passes over the 45° line, the "nearest edge point" would
	//	jump suddenly from one edge to the other. That jump
	//	in the basepoint for the tangent space would cause
	//	a small -- but potentially perceptible -- jump in the headset's
	//	position in the virtual hyperbolic or spherical space.
	//
	//	To avoid the discontinuous jump, we can base the tangent space
	//	not at the nearest point on the nearest table edge, but rather
	//	at the table edge point whose azimuth agrees with the azimuth
	//	of the player's head, as seen from the center of the table.
	//
	//	The above description defines the mapping (from physical space
	//	to virtual space) almost uniquely.  The only "free parameter"
	//	is deciding which edge of the physical table corresponds
	//	to which edge of the virtual table. The corresondece between
	//	the edges of the square physical table and the edges of
	//	the triangular, square or pentagonal table is not fixed,
	//	but is something that must be traced around in real time,
	//	as the player walks around the table.  For example,
	//	if a player keeps walking around the physical table -- always
	//	in the same direction, never turning back -- s/he'll see
	//	the following edges in the following order.
	//
	//	  square physical table:     0 1 2 3 0 1 2 3 0 1 2 3 0 1 2 …
	//	  pentagonal virtual table:  0 1 2 3 4 0 1 2 3 4 0 1 2 3 4 …
	//
	//	updateNearbyTableEdgeIndices() handles this correspondence.
	//
	//
	//	Note: A related question concerns the relationship between
	//	the player's head and hands. For example, if you're
	//	in a simulation of hyperbolic 3-space and you move your head
	//	around in circles, your head would end up slightly rotated
	//	relative to your body in hyperbolic space, due to "holonomy".
	//	But of course in the physical room this doesn't happen.
	//	So what you'd see with your eyes would disagree with what
	//	you feel with your body.  In particular, you'd see your hands
	//	floating in space at a different place than where you feel them to be.
	//	If one could visit a real hyperbolic space, would this happen?
	//	No, of course not.  Instead, one would need to rotate one's
	//	head as it orbits around. (Note: Contrary to my original thoughts
	//	on the question, I no longer think that you'd feel a torque
	//	in your neck the whole time. Instead, once you started your
	//	head rotating, it could, it seems, keep rotating even with
	//	no further torque applied.) As a practical matter, the app
	//	makes sure that the player's virtual hands always maintain
	//	the "same" pose relative to the player's virtual head that
	//	his/her physical hands maintain relative to his/her physical head.
	//
	//
	//	In summary:
	//
	//		While this Billiards app models the table,
	//			the balls and the balls' motions *exactly*,
	//
	//		it does *not* model the player's motions exactly,
	//			but instead contrains them to keep the player's
	//			body (head and hands) coherent relative to the table
	//			and to each other.
	//
	
	let spaceGeometry: NEBSpaceGeometry

	//	As the player walks around the table, keep track
	//	of which edge of the implicit physical table
	//	and which edge of the virtual table is nearby.
	var nearbyPhysicalTableEdgeIndex = 0	//	∈ {0,1,2,3}, because the implicit physical table is square
	var nearbyVirtualTableEdgeIndex = 0		//	∈ {0,1,2}, {0,1,2,3} or {0,1,2,3,4},
											//	according to whether the virtual table is
											//		triangular (spherical geometry),
											//		square     (Euclidean geometry) or
											//		pentagonal (hyperbolic geometry).

	init(
		spaceGeometry: NEBSpaceGeometry
	) {
		self.spaceGeometry = spaceGeometry
	}

	func getVirtualHeadsetPose(
		physicalPose: simd_double4x4	//	headset pose in table coordinates
										//		(not world coordinates)
	) -> simd_double4x4 {	//	returns virtual pose as 4×4 matrix (dimensionless),
							//	also in table coordinates

		//	updateNearbyTableEdgeIndices() chooses a basepoint
		//	whose azimuth agrees with the headset's.
		//	getTheVirtualHeadsetPose() then uses that basepoint.
		
		let theBasepointDistanceAlongEdge
			= updateNearbyTableEdgeIndices(
				physicalHeadsetPose: physicalPose)
		
		let theVirtualHeadsetPose
			= getTheVirtualHeadsetPose(
				physicalPose: physicalPose,
				basepointDistanceAlongEdge: theBasepointDistanceAlongEdge)
		
		return theVirtualHeadsetPose
	}
	
	func updateNearbyTableEdgeIndices(
		physicalHeadsetPose: simd_double4x4	//	in table coordinates (not world coordinates)
	) -> Double {	//	returns the basepoint's distance along the edge, in meters
	
		//	Whenever the player walks past a corner
		//	of the implicit physical table, either increment both
		//
		//		nearbyPhysicalTableEdgeIndex
		//		nearbyVirtualTableEdgeIndex
		//
		//	or decrement both, according to whether the player
		//	is walking counterclockwise or clockwise.
		//	In all cases the indices are treated cyclically.

		//	In the zx plane, the implicit physical table's edges
		//	are indexed like this:
		//
		//		          -Z
		//		           |
		//		      22222222222
		//		     3     |     1
		//		     3     |     1
		//		-X --3-----+-----1-- +X
		//		     3     |     1
		//		     3     |     1
		//		      00000000000
		//		           |
		//		          +Z
		//
		//	The virtual table edges are indexed similarly,
		//	with edge 0 always centered on the +z axis.
		
		//	How many sides does the virtual table have?
		let n = spaceGeometry.geometryType.tableNumSides

		//	Which edge of the implicit physical table is currently
		//	nearest to the player?
		
		let theAzimuth = atan2(	//	= atan(x/z) but in range [-π, +π];  atan2(0.0, 0.0) = 0.0
			physicalHeadsetPose[3][0],	//	x coordinate
			physicalHeadsetPose[3][2])	//	z coordinate

		//	Caution
		//
		//	'%' is a "remainder" operator, not a "modulus" operator,
		//	so throughout the following computations we must be careful
		//	to apply it to positive numbers only.

		let theCurrentNearbyPhysicalTableEdgeIndex
			= (Int(floor((theAzimuth / (0.5*Double.pi)) + 0.5)) + 4) % 4

		//	Has the player moved to a different edge?
		switch theCurrentNearbyPhysicalTableEdgeIndex - nearbyPhysicalTableEdgeIndex {
		
		case 0:
		
			//	The player is near the same edge as s/he was during the previous frame.
			//	No changes are needed.  This is the typical case.
			break
		
		case +1, -3:
		
			//	The player has moved to the next higher numbered edge.
			nearbyPhysicalTableEdgeIndex = (nearbyPhysicalTableEdgeIndex + 1) % 4
			nearbyVirtualTableEdgeIndex  = (nearbyVirtualTableEdgeIndex  + 1) % n
		
		case -1, +3:
		
			//	The player has moved to the next lower numbered edge.
			nearbyPhysicalTableEdgeIndex = (nearbyPhysicalTableEdgeIndex + (4 - 1)) % 4
			nearbyVirtualTableEdgeIndex  = (nearbyVirtualTableEdgeIndex  + (n - 1)) % n
		
		default:	//	-2, +2

			//	The player has somehow jumped across the middle of the table
			//	to the opposite side, without passing by either of the
			//	intervening edges.  Either that or there's some bizarre error --
			//	maybe the Vision Pro's cameras were blocked for a few seconds.
			//	My hope is that this case will never arise, but if it does,
			//	let's just update nearbyPhysicalTableEdgeIndex and leave
			//	nearbyVirtualTableEdgeIndex unchanged.  The player
			//	will see a discontinuous jump in the scene, but such a jump
			//	may be unavoidable in this bizarre case.

			nearbyPhysicalTableEdgeIndex = theCurrentNearbyPhysicalTableEdgeIndex
		}

		//	For simplicity, rotate the physical headset pose
		//	so that it sits near "edge 0", which runs from
		//
		//		(-thePhysicalTableInradius, 0.0, thePhysicalTableInradius)
		//	to
		//		(+thePhysicalTableInradius, 0.0, thePhysicalTableInradius)
		//
		let θ = Double(nearbyPhysicalTableEdgeIndex) * (2.0 * Double.pi / 4.0)
		let theSimplicityRotation = simd_double4x4(
			SIMD4<Double>( cos(θ), 0.0,  sin(θ), 0.0),
			SIMD4<Double>(  0.0,   1.0,   0.0,   0.0),
			SIMD4<Double>(-sin(θ), 0.0,  cos(θ), 0.0),
			SIMD4<Double>(  0.0,   0.0,   0.0,   1.0)
		)
		let theRotatedPhysicalHeadsetPose = theSimplicityRotation * physicalHeadsetPose

		//	On the nearby physical table edge, which is now edge 0,
		//	locate the basepoint whose azimuth agrees with the azimuth
		//	of theRotatedPhysicalPose, as seen from the center of the table.
		//	The simplest way to compute this basepoint is to notice
		//	that its z coordinate must equal the table's inradius,
		//	and then scale the x coordinate proportionally.

		//	Note:  ARKit implicitly models physical space as a Euclidean space
		//	of radius 1 meter (because each point's w coordinate is 1).
		//
		let thePhysicalX = 1.0 * theRotatedPhysicalHeadsetPose[3][0]	//	(1 meter)*(m₃₀)
		let thePhysicalZ = 1.0 * theRotatedPhysicalHeadsetPose[3][2]	//	(1 meter)*(m₃₂)

		let thePhysicalTableInradius
			= 0.5 * spaceGeometry.tableEdgeLengthInMeters	//	always a square table

		let theTableCenterEpsilon = 0.01	//	1 cm inradius of square containing table center

		let theBasepointDistanceAlongEdge = (
		
			thePhysicalZ > theTableCenterEpsilon
			?
				//	Typical case
				thePhysicalX * (thePhysicalTableInradius / thePhysicalZ)
			:
				//	Atypical case -- player's head is near table center.
				//	This will cause a discontinuous jump or shudder.
				//	Such behavior near the table center is inevitable.
				0.0
		)
		
		return theBasepointDistanceAlongEdge
	}

	func getTheVirtualHeadsetPose(
		physicalPose: simd_double4x4,	//	headset pose in table coordinates
										//		(not world coordinates)
		basepointDistanceAlongEdge: Double
	) -> simd_double4x4 {	//	returns virtual headset pose as 4×4 matrix (dimensionless),
							//	also in table coordinates

		//	Map the pose from the Euclidean physical space
		//	into the potentially non-Euclidean virtual space.

		//	Some basic parameters
		let n = spaceGeometry.geometryType.tableNumSides
		let thePhysicalTableInradius
			= 0.5 * spaceGeometry.tableEdgeLengthInMeters	//	always a square table

		//	Recall from updateNearbyTableEdgeIndices() that, in the zx plane,
		//	the implicit physical table's edges are indexed like this:
		//
		//		          -Z
		//		           |
		//		      22222222222
		//		     3     |     1
		//		     3     |     1
		//		-X --3-----+-----1-- +X
		//		     3     |     1
		//		     3     |     1
		//		      00000000000
		//		           |
		//		          +Z
		//
		//	For simplicity, rotate the physical pose
		//	so that it sits near "edge 0", which runs from
		//
		//		(-thePhysicalTableInradius, 0.0, thePhysicalTableInradius)
		//	to
		//		(+thePhysicalTableInradius, 0.0, thePhysicalTableInradius)
		//
		let θ = Double(nearbyPhysicalTableEdgeIndex) * (2.0 * Double.pi / 4.0)
		let theSimplicityRotation = simd_double4x4(
			SIMD4<Double>( cos(θ), 0.0,  sin(θ), 0.0),
			SIMD4<Double>(  0.0,   1.0,   0.0,   0.0),
			SIMD4<Double>(-sin(θ), 0.0,  cos(θ), 0.0),
			SIMD4<Double>(  0.0,   0.0,   0.0,   1.0)
		)
		let theRotatedPhysicalPose = theSimplicityRotation * physicalPose

		//	On the nearby physical table edge, which is now edge 0,
		//	updateNearbyTableEdgeIndices() has already chosen
		//	a suitable basepoint at
		//
		//	  (
		//		basepointDistanceAlongEdge,
		//		0.0,
		//		thePhysicalTableInradius,
		//		1.0   	(the physical "space radius" is always 1 meter)
		//	  )
		//

		//	Factor theRotatedPhysicalPose as a rotation followed by two translations.

		let thePhysicalRotation = simd_double4x4(
			theRotatedPhysicalPose[0],
			theRotatedPhysicalPose[1],
			theRotatedPhysicalPose[2],
			SIMD4<Double>(0.0, 0.0, 0.0, 1.0)
		)
		let thePhysicalOriginToBasepointTranslation = SIMD3<Double>(
			basepointDistanceAlongEdge,
			0.0,
			thePhysicalTableInradius
		)
		let thePhysicalBasepointToEntityTranslation = SIMD3<Double>(
			theRotatedPhysicalPose[3][0] - thePhysicalOriginToBasepointTranslation.x,
			theRotatedPhysicalPose[3][1] - thePhysicalOriginToBasepointTranslation.y,
			theRotatedPhysicalPose[3][2] - thePhysicalOriginToBasepointTranslation.z
		)

		//	Use those same rotation and translations to compute theVirtualPose,
		//	no longer working in the context of a flat physical space of radius 1 meter,
		//	but instead in a curved virtual space of radius spaceGeometry.spaceRadius.

		//	thePhysicalRotation is a pure rotation in the xyz hyperplane,
		//	with matrix
		//
		//			( m₀₀ m₀₁ m₀₂ 0 )
		//			( m₁₀ m₁₁ m₁₂ 0 )
		//			( m₂₀ m₂₁ m₂₂ 0 )
		//			(  0   0   0  1 )
		//
		//	Even though we constructed this matrix as a rotation
		//	of the flat physical space (with "space radius" 1 meter),
		//	we may re-interpret it as a rotation of the possibly curved
		//	virtual space (with space radius spaceGeometry.spaceRadius).
		let theVirtualRotation = thePhysicalRotation

		//	Let the basepoint-to-entity translation be
		//	a single translation along a geodesic, which ensures
		//	that the final physical pose will
		//	vary smoothly even when the basepoint passes
		//	from one of the table's edges to the next
		//	(in other words, when the basepoint passes through a corner).
		let theVirtualBasepointToEntityTranslation
			= geodesicTranslation(
				v: thePhysicalBasepointToEntityTranslation / spaceGeometry.spaceRadius,
				geometryType: spaceGeometry.geometryType)

		//	The origin-to-basepoint translation gets applied
		//	one component at a time, to ensure that ultimately
		//	an orthonormal frame aligned with an edge of the implicit
		//	physical table will map to an orthonormal frame aligned
		//	with the corresonding edge of the virtual table.
		//
		//		Note #1:  While the virtual edge length in meters
		//		is the same as the physical edge length in meters,
		//		the space radii are different and thus the edge lengths
		//		in radians are also different.
		//
		//		Note #2:  The inradii are different too.
		//

		let theXTranslation
			= axisAlignedTranslation(
				axis: 0,
				distanceInRadians: basepointDistanceAlongEdge / spaceGeometry.spaceRadius,
				geometryType: spaceGeometry.geometryType)

	//	let theYTranslation
	//		= axisAlignedTranslation(
	//			axis: 1,
	//			distanceInRadians: 0.0 / spaceGeometry.spaceRadius,
	//			geometryType: spaceGeometry.geometryType)

		let theZTranslation
			= axisAlignedTranslation(
				axis: 2,
				distanceInRadians: spaceGeometry.geometryType.tableInradius,
				geometryType: spaceGeometry.geometryType)

		//	Apply the translation components in the order
		//
		//		X	horizontal translation parallel to edge
		//		Z	horizontal translation perpendicular to edge
		//		Y	vertical translation (which is zero)
		//
		let theVirtualOriginToBasepointTranslation
			= theZTranslation * theXTranslation	//	right-to-left matrix composition

		//	Rotate the table edge from its current position on the positive z axis
		//	to the desired position on the virtual table.
		//
		//	  Note:
		//		The following two lines and the code in makeTableMesh()
		//		and cushionNormalVectorsXZW() all share the assumptions
		//		that edge 0 sits with its midpoint on the positive z axis,
		//		and that the remaining edges progress around in the direction
		//		of the positive x axis.
		//
		let φ = Double(nearbyVirtualTableEdgeIndex) * (2.0 * Double.pi / Double(n))
		let theTableRotation = simd_double4x4(
			SIMD4<Double>( cos(φ), 0.0, -sin(φ), 0.0),
			SIMD4<Double>(  0.0,   1.0,   0.0,   0.0),
			SIMD4<Double>( sin(φ), 0.0,  cos(φ), 0.0),
			SIMD4<Double>(  0.0,   0.0,   0.0,   1.0)
		)

		//	Multiply together the four factors.
		let theVirtualEntityPose	//	right-to-left matrix composition
			= theTableRotation
			* theVirtualOriginToBasepointTranslation
			* theVirtualBasepointToEntityTranslation
			* theVirtualRotation

		return theVirtualEntityPose
	}
}
