//	NEBRenderer.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import CompositorServices
import MetalKit	//	for MTKTextureLoader


let groundingShadowMeshRefinement: Int = 32
let pocketMeshRefinement: Int = 64


actor NEBRenderer {

	let layerRenderer: LayerRenderer

	let playerHandedness: PlayerHandedness
	let computerPlayerDifficulty: Double	//	∈ 0.0 ... 1.0
	let adjustPerceivedDistances: Bool
	let tableHeight: Double
	
	//	When the player begins a game, s/he must
	//
	//		walk backwards past the edge of the table,
	//			at which point we'll give him/her a cue ball,
	//
	//		raise the cue ball above the height of the table, and
	//
	//		place the cue ball onto the table.
	//
	//	When the Control Panel is dismissed -- either because
	//	the player explicitly dismissed it after reading
	//	the optional hints panels or because it got automatically
	//	dismissed with the hints disabled -- we'll give the player
	//	a cue stick and let the game begin.
	//
	//		Note: As each of the above conditions is satisfied
	//		we'll set the corresponding flag to true.
	//		But we'll never set any of these flags back to false.
	//		It's enough that the condition has happened once.
	//
	var playerHasSteppedBackFromTable: Bool = false
	var playerHasRaisedCueBallAboveTable: Bool = false
	var playerHasPlacedCueBallOnTable: Bool = false
	var instructionsPanelIsVisible: Bool = true	//	assume panel is visible when immersive space opens
	
	//	As the player progresses through the starting sequence
	//	described immediately above, we'll enable
	//	first the cue ball and then the cue stick.
	var cueBallIsAvailable: Bool = false
	var cueStickIsAvailable: Bool = false

	var playerSteppedBack: @Sendable () -> Void
	var playerRaisedCueBall: @Sendable () -> Void
	var playerPlacedCueBall: @Sendable () -> Void
	var playerClosedImmersiveSpace: @Sendable () -> Void
	var temporaryClosure: @Sendable () -> Void

	//	We'll access the gameState only from our render loop,
	//	so it should be thread-safe even with no locks.
	var gameState: NEBGameState
	
	var previousPresentationTime: TimeInterval?

	let arKitSession: ARKitSession
	let worldTrackingProvider: WorldTrackingProvider
	let handTrackingProvider: HandTrackingProvider
#warning("Delete playersHandWasAlreadyRaised in some future release, when players are used to pressing the Digital Crown to close the immersive space.")
	var playersHandWasAlreadyRaised: Bool = false	//	…during previous pass through render loop
	var poseMapper: NEBPoseMapper

	let computerPlayer: NEBComputerPlayer
	
	//	I'm slightly uncomfortable using implicitly unwrapped optionals
	//	-- I hadn't intended to use them at all, for run-time safety --
	//	but in this case I really would rather not have to get
	//	all the Metal objects set up directly in init(),
	//	without being able to make any function calls.
	//	So let's go ahead and use implicitly unwrapped optionals here.
	
	var device: MTLDevice!
	var commandQueue: MTLCommandQueue!
	
	var billiardsTablePipelineState: MTLRenderPipelineState!
	var groundingShadowPipelineState: MTLRenderPipelineState!
	var pocketPipelineState: MTLRenderPipelineState!
	var billiardBallPipelineState: MTLRenderPipelineState!
	var cueStickPipelineState: MTLRenderPipelineState!
	var handPipelineState: MTLRenderPipelineState!

	var enabledDepthStencilState: MTLDepthStencilState!
	var disabledDepthStencilState: MTLDepthStencilState!
	
	var tableMesh: NEBMesh!
	var groundingShadowMesh: NEBMesh!
	var pocketMesh: NEBMesh!
	var ballMesh: NEBMesh!
	var cueStickMesh: NEBMesh!
	var handFacets: MTLBuffer!
	
	var viewProjectionBufferPool: GGVRBufferPool!

	var ballTextures: [MTLTexture] = []
	var cueStickTexture: MTLTexture!
	
	var ballSampler: MTLSamplerState!
	var cueStickSampler: MTLSamplerState!

	enum PlayersHand: CaseIterable {
		case left
		case right
	}


// MARK: -
// MARK: Init

	init?(
		layerRenderer: LayerRenderer,
		playerHandedness: PlayerHandedness,
		computerPlayerDifficulty: Double,	//	∈ 0.0 ... 1.0
		guardianAngelFactor: Double,		//	∈ 0.0 ... 1.0
		adjustPerceivedDistances: Bool,
		geometryType: GeometryType,
		tableEdgeLengthInMeters: Double,
		tableHeight: Double,
		playerSteppedBack: @Sendable @escaping () -> Void,
		playerRaisedCueBall: @Sendable @escaping () -> Void,
		playerPlacedCueBall: @Sendable @escaping () -> Void,
		playerClosedImmersiveSpace: @Sendable @escaping () -> Void,
		temporaryClosure: @Sendable @escaping () -> Void
	) async {

		self.layerRenderer = layerRenderer
		self.playerHandedness = playerHandedness
		self.computerPlayerDifficulty = computerPlayerDifficulty
		self.adjustPerceivedDistances = adjustPerceivedDistances
		self.tableHeight = tableHeight

		let theSpaceGeometry = NEBSpaceGeometry(
								geometryType: geometryType,
								tableEdgeLengthInMeters: tableEdgeLengthInMeters)

		self.gameState = NEBGameState(
							spaceGeometry: theSpaceGeometry,
							guardianAngelFactor: guardianAngelFactor)

		self.poseMapper = NEBPoseMapper(spaceGeometry: theSpaceGeometry)
		self.computerPlayer = NEBComputerPlayer()

		self.playerSteppedBack = playerSteppedBack
		self.playerRaisedCueBall = playerRaisedCueBall
		self.playerPlacedCueBall = playerPlacedCueBall
		self.playerClosedImmersiveSpace = playerClosedImmersiveSpace
		self.temporaryClosure = temporaryClosure
		
		self.arKitSession = ARKitSession()
		self.worldTrackingProvider =  WorldTrackingProvider()
		self.handTrackingProvider = HandTrackingProvider()

		if !setUpMetalObjects() {
			return nil
		}
	}
	
	deinit {
		playerClosedImmersiveSpace()
	}

// MARK: -
// MARK: World and hand tracking

	func startWorldAndHandTracking() async {
	
		do {
#if targetEnvironment(simulator)
			//	The simulator doesn't support hand tracking.
			try await arKitSession.run([worldTrackingProvider])
#else
			try await arKitSession.run([worldTrackingProvider, handTrackingProvider])
#endif
		} catch {
			preconditionFailure("ARKitSession error: \(error)")
		}
	}


// MARK: -
// MARK: Set up Metal objects

	func setUpMetalObjects(
	) -> Bool {
		
		device = layerRenderer.device
		
		guard let theCommandQueue = device.makeCommandQueue()
		else { return false }
		commandQueue = theCommandQueue

		if  setUpPipelineStates()
		 && setUpDepthStencilStates()
		 && setUpFixedBuffers()
		 && setUpInflightBufferPools()
		 && setUpTextures()
		 && setUpSamplerStates()
		{
			return true
			
		} else {
		
			assertionFailure("Unable to set up Metal resources")
			return false
		}
	}

	func setUpPipelineStates() -> Bool {

		guard let theGPUFunctionLibrary = device.makeDefaultLibrary() else {
			assertionFailure("Failed to create default GPU function library")
			return false
		}

		let theCompileTimeConstants = MTLFunctionConstantValues()
		
		var adjustPerceivedDistancesAsVar: Bool = adjustPerceivedDistances
		theCompileTimeConstants.setConstantValue(
			&adjustPerceivedDistancesAsVar,
			type: .bool,
			withName: "gAdjustPerceivedDistances")
		
		var geometryCode: UInt8 = gameState.spaceGeometry.geometryType.gpuGeometryCode
		theCompileTimeConstants.setConstantValue(
			&geometryCode,
			type: .uchar,
			withName: "gGeometryCode")
		
		var spaceRadiusAsVar: Float32 = Float32(gameState.spaceGeometry.spaceRadius)
		theCompileTimeConstants.setConstantValue(
			&spaceRadiusAsVar,
			type: .float,
			withName: "gSpaceRadius")
			
		var falseValue: Bool = false
		var trueValue: Bool = true


		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gBilliardBallShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gCueStickShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gAntiflickering")
		let theNonShadingVertexFunction: MTLFunction
		do {
			theNonShadingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theNonShadingVertexFunction: \(error)")
			return false
		}
		theNonShadingVertexFunction.label = "non-shading vertex function"

		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gBilliardBallShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gCueStickShading")
		theCompileTimeConstants.setConstantValue(
			&trueValue,
			type: .bool,
			withName: "gAntiflickering")
		let theBilliardsTableVertexFunction: MTLFunction
		do {
			theBilliardsTableVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theBilliardsTableVertexFunction: \(error)")
			return false
		}
		theBilliardsTableVertexFunction.label = "billiards table vertex function"

		//	Apple explicitly says it's OK to modify and reuse MTLFunctionConstantValues.
		theCompileTimeConstants.setConstantValue(
			&trueValue,
			type: .bool,
			withName: "gBilliardBallShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gCueStickShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gAntiflickering")
		let theBallShadingVertexFunction: MTLFunction
		do {
			theBallShadingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theBallShadingVertexFunction: \(error)")
			return false
		}
		theBallShadingVertexFunction.label = "ball-shading vertex function"

		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gBilliardBallShading")
		theCompileTimeConstants.setConstantValue(
			&trueValue,
			type: .bool,
			withName: "gCueStickShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gAntiflickering")
		let theCueStickShadingVertexFunction: MTLFunction
		do {
			theCueStickShadingVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theCueStickShadingVertexFunction: \(error)")
			return false
		}
		theCueStickShadingVertexFunction.label
			= "cue stick shading vertex function"

		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gBilliardBallShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gCueStickShading")
		theCompileTimeConstants.setConstantValue(
			&falseValue,
			type: .bool,
			withName: "gAntiflickering")
		let theHandVertexFunction: MTLFunction
		do {
			theHandVertexFunction = try theGPUFunctionLibrary.makeFunction(
				name: "nonEuclideanBilliardsVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure("Failed to load theHandVertexFunction: \(error)")
			return false
		}
		theHandVertexFunction.label = "hand vertex function"

		guard let theBilliardsTableFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "billiardsTableFragmentFunction")
		else {
			assertionFailure("Failed to load billiardsTableFragmentFunction")
			return false
		}
		theBilliardsTableFragmentFunction.label = "billiards table fragment function"

		guard let theGroundingShadowFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "groundingShadowFragmentFunction")
		else {
			assertionFailure("Failed to load groundingShadowFragmentFunction")
			return false
		}
		theGroundingShadowFragmentFunction.label = "grounding shadow fragment function"

		guard let thePocketFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "pocketFragmentFunction")
		else {
			assertionFailure("Failed to load pocketFragmentFunction")
			return false
		}
		thePocketFragmentFunction.label = "pocket fragment function"

		guard let theBilliardBallFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "billiardBallFragmentFunction")
		else {
			assertionFailure("Failed to load billiardBallFragmentFunction")
			return false
		}
		theBilliardBallFragmentFunction.label = "billiard ball fragment function"

		guard let theCueStickFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "cueStickFragmentFunction")
		else {
			assertionFailure("Failed to load cueStickFragmentFunction")
			return false
		}
		theCueStickFragmentFunction.label = "cue stick fragment function"

		guard let theHandFragmentFunction
			= theGPUFunctionLibrary.makeFunction(name: "handFragmentFunction")
		else {
			assertionFailure("Failed to load theHandFragmentFunction")
			return false
		}
		theHandFragmentFunction.label = "hand fragment function"

		//	Create a MTLVertexDescriptor describing the vertex attributes
		//	that the GPU sends along to each vertex.  Each vertex attribute
		//	may be a 1-, 2-, 3- or 4-component vector.
		
		let theVertexDescriptor = MTLVertexDescriptor()
		
			//	Say where to find each attribute.

		guard let thePosOffset = MemoryLayout.offset(of: \NEBVertex.pos) else {
			assertionFailure("Failed to find offset of NEBVertex.pos")
			return false
		}
		theVertexDescriptor.attributes[VertexAttributePosition].format = MTLVertexFormat.float4
		theVertexDescriptor.attributes[VertexAttributePosition].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributePosition].offset = thePosOffset

		guard let theTexOffset = MemoryLayout.offset(of: \NEBVertex.tex) else {
			assertionFailure("Failed to find offset of NEBVertex.tex")
			return false
		}
		theVertexDescriptor.attributes[VertexAttributeTexCoords].format = MTLVertexFormat.float3
		theVertexDescriptor.attributes[VertexAttributeTexCoords].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributeTexCoords].offset = theTexOffset

			//	Say how to step through each buffer.

		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stepFunction = MTLVertexStepFunction.perVertex
		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stride = MemoryLayout<NEBVertex>.stride

		//	Create the pipeline states.

		guard let theBilliardsTablePipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theBilliardsTableVertexFunction,
				fragmentFunction: theBilliardsTableFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		billiardsTablePipelineState = theBilliardsTablePipelineState
		
		guard let theGroundingShadowPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theNonShadingVertexFunction,
				fragmentFunction: theGroundingShadowFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: true)
		else { return false }
		groundingShadowPipelineState = theGroundingShadowPipelineState

		guard let thePocketPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theNonShadingVertexFunction,
				fragmentFunction: thePocketFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		pocketPipelineState = thePocketPipelineState
		
		guard let theBilliardBallPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theBallShadingVertexFunction,
				fragmentFunction: theBilliardBallFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		billiardBallPipelineState = theBilliardBallPipelineState

		guard let theCueStickPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theCueStickShadingVertexFunction,
				fragmentFunction: theCueStickFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		cueStickPipelineState = theCueStickPipelineState
		
		guard let theHandPipelineState = ggvrPipelineState(
				layerRenderer: layerRenderer,
				vertexFunction: theHandVertexFunction,
				fragmentFunction: theHandFragmentFunction,
				vertexDescriptor: theVertexDescriptor,
				enableBlending: false)
		else { return false }
		handPipelineState = theHandPipelineState

		return true
	}
		
	func setUpDepthStencilStates() -> Bool {

		if let theEnabledDepthStencilState = ggvrDepthStencilState(
												device: device,
												depthTestingEnabled: true)
		{
			enabledDepthStencilState = theEnabledDepthStencilState
		} else {
			return false
		}

		if let theDisabledDepthStencilState = ggvrDepthStencilState(
												device: device,
												depthTestingEnabled: false)
		{
			disabledDepthStencilState = theDisabledDepthStencilState
		} else {
			return false
		}

		return true
	}
	
	func setUpFixedBuffers() -> Bool {
	
		tableMesh = makeTableMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)

		groundingShadowMesh = makeDiskMesh(
						device: device,
						r: gameState.spaceGeometry.ballShadowRadius,
						n: groundingShadowMeshRefinement,
						spaceGeometry: gameState.spaceGeometry)

		pocketMesh = makeDiskMesh(
						device: device,
						r: pocketRadius,
						n: pocketMeshRefinement,
						spaceGeometry: gameState.spaceGeometry)

		ballMesh = makeBallMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)
		
		cueStickMesh = makeCueStickMesh(
						device: device,
						spaceGeometry: gameState.spaceGeometry)
		
		handFacets = ggvrMeshFacets(
						device: device,
						vertexIndices: gHandMeshFacets)
		
		return true
	}
	
	func setUpInflightBufferPools() -> Bool {
	
		viewProjectionBufferPool = GGVRBufferPool(
			device: device,
			initialBufferSize: MemoryLayout<GGVRViewProjection>.stride,
			storageMode: .storageModeShared,
			bufferLabel: "View-projection buffer")
		
		return true
	}
	
	func setUpTextures() -> Bool {
	
		let theTextureLoader = MTKTextureLoader(device: device)
		let theTextureLoaderOptions = [
		
			MTKTextureLoader.Option.textureUsage
				: NSNumber(value: MTLTextureUsage.shaderRead.rawValue),
				
			MTKTextureLoader.Option.textureStorageMode
				: NSNumber(value: MTLStorageMode.`private`.rawValue),
			
			//	Mipmaps
			//
			//	Our Asset Catalog automatically generates mipmaps
			//	for all textures (at compile time? at run time? I'm not sure).
			//	To see the mipmap setting, go to NEBTextures.xcassets,
			//	open the Xcode's right-hand sidebar, and select
			//	the last tab (the Attributes inspector).  Then select
			//	one of the textures, for example
			//
			//		cue stick > Universal
			//
			//	and look in the right-hand side bar to see
			//	its Mipmap Set options.  Note that you have to select
			//	a specific image -- here "Universal" -- for this to work;
			//	if you'd instead just selected "cue stick" you wouldn't
			//	see the Mipmap Set options.
			//
			//	Regarding MTKTextureLoader's mipmap-related options...
			//	Apple's page
			//
			//		https://developer.apple.com/documentation/metalkit/mtktextureloader/option/1535977-allocatemipmaps
			//
			//	says that
			//
			//		[The allocateMipmaps] option applies only if the texture
			//		being loaded does not contain mipmap data.
			//		...
			//		If this key is not specified and the image being loaded
			//		contains data for mipmaps, the mipmap memory is allocated
			//		and the image data is loaded.
			//
			//	so we can omit
			//
			//		MTKTextureLoader.Option.allocateMipmaps : NSNumber(value: true),
			//
			//	Similarly, Apple's page
			//
			//		https://developer.apple.com/documentation/metalkit/mtktextureloader/1645851-newtexture
			//
			//	says that
			//
			//		"When using [newTexture()], the texture loader
			//		ignores the generateMipmaps … option."
			//
			//	Instead, the Asset Catalog provides the mipmaps.
			//	So we can omit
			//
			//		MTKTextureLoader.Option.generateMipmaps : NSNumber(value: true)
			//
		]
		
		//	Cube map coordinate conventions
		//
		//	Apple's MTLTexture documentation says which index (in 0...5)
		//	maps to which face of the cube, but doesn't say how it maps
		//	to that face.  In practice the layout seems to be
		//
		//
		//		       +------+                   +---→ s
		//		       |      |                   |
		//		       |  y+  |                   ↓ t
		//		       |      |
		//		+------+------+------+------+
		//		|      |      |      |      |
		//		|  x-  |  z+  |  x+  |  z-  |
		//		|      |      |      |      |
		//		+------+------+------+------+
		//		       |      |
		//		       |  y-  |
		//		       |      |
		//		       +------+
		//
		//
		//	with each face oriented the same as in the old
		//	OpenGL ARB_texture_cube_map extension, namely
		//
		//				   largest
		//		  index     coord   pre-s  pre-t
		//		  -----    -------  -----  -----
		//			0       x > 0    -z     -y
		//			1       x < 0    +z     -y
		//			2       y > 0    +x     +z
		//			3       y < 0    +x     -z
		//			4       z > 0    +x     -y
		//			5       z < 0    -x     -y
		//
		//	with the final s and t texture coordinates given by
		//
		//		s = ( ( pre-s / largest-coord ) + 1 ) / 2
		//		t = ( ( pre-t / largest-coord ) + 1 ) / 2
		//
		//	These conventions agree with the diagram shown above.
		//
		//	Note:  An Xcode Asset Catalog's images seem to be
		//	upside down relative to the conventions just described.
		//	Most likely this is just a question of whether
		//	the image rows are being read top-down or bottom-up.
		//
		ballTextures = []
		do {
			for i in 0 ..< numBalls {
				let theBallTexture = try theTextureLoader.newTexture(
										name: "\(i) ball",
										scaleFactor: 1.0,
										bundle: nil,
										options: theTextureLoaderOptions)
				ballTextures.append(theBallTexture)
			}
		} catch {
			assertionFailure("Failed to load one or more ball textures.  Error:  \(error)")
			return false
		}
	
		do {
			cueStickTexture = try theTextureLoader.newTexture(
									name: "cue stick",
									scaleFactor: 1.0,
									bundle: nil,
									options: theTextureLoaderOptions)
		} catch {
			assertionFailure("Failed to load cue stick textures.  Error:  \(error)")
			return false
		}
		
		return true
	}
	
	func setUpSamplerStates() -> Bool {

		//	Presumably the MTLSamplerAddressModes are irrelevant with a cube map?
		guard let theBallSampler = ggvrSamplerState(
									device: device,
									sMode: .repeat,
									tMode: .repeat)
		else {
			assertionFailure("Couldn't create theBallSampler")
			return false
		}
		ballSampler = theBallSampler
		
		//	tMode:.repeat is essential to accommodate meridional phase shift.
		guard let theCueStickSampler = ggvrSamplerState(
										device: device,
										sMode: .clampToEdge,
										tMode: .repeat)
		else {
			assertionFailure("Couldn't create cueStickSampler")
			return false
		}
		cueStickSampler = theCueStickSampler
		
		return true
	}


// MARK: -
// MARK: Render loop

	func runRenderLoop() {

		var keepRendering = true	//	remains true until the immersive space gets closed

		while keepRendering {
		
			switch layerRenderer.state {
			
			case .paused:
				layerRenderer.waitUntilRunning()
				
			case .running:
				autoreleasepool {
					renderOneFrame()
				}
				
			case .invalidated:
				fallthrough
			@unknown default:
				keepRendering = false
			}
		}
		
		arKitSession.stop()
	}

	func renderOneFrame() {
	
		if instructionsPanelIsVisible && gInstructionsPanelHasClosed.get() {
			instructionsPanelIsVisible = false
		}
		if cueBallIsAvailable && !instructionsPanelIsVisible {
			cueStickIsAvailable = true
		}

		guard let theFrame = layerRenderer.queryNextFrame(),
			  let theTiming = theFrame.predictTiming()
		else {
			assertionFailure("Couldn't get theFrame and/or theTiming")
			return
		}
		
// MARK: -
// MARK: "Update" section

		//	Update anything that doesn't depend on the headset position.
		//	The startUpdate() / endUpdate() calls help
		//	Compositor Services plan its timing for future frames.

		theFrame.startUpdate()

		//	Compute the frame interval, which should be
		//	1/90 sec on a Vision Pro or 1/60 sec in the simulator.
		let theFrameInterval: Double
		let thePredictedPresentationTime = LayerRenderer.Clock.Instant.epoch.duration(
											to: theTiming.presentationTime).timeInterval
		if let thePreviousPresentationTime = previousPresentationTime {
			theFrameInterval = thePredictedPresentationTime
							 - thePreviousPresentationTime
		} else {
			theFrameInterval = 1.0/90.0	//	should occur only once, on the very first frame
		}
		previousPresentationTime = thePredictedPresentationTime	//	non-nil

		//	Did either we or our computer opponent just now take a shot?

		let theCueStick = gameState.cueStick[gameState.whoseTurn]!
		
		_ = gameState.maybeTakeShot(
				cueStick: theCueStick,
				handTrackingInterval: theFrameInterval)

		//	The same code that lets the computer take a shot also serves
		//	to let our "computer self" play when running in the simulator,
		//	so let's enclose it in a function that can be called in both situations.
		func maybeLetComputerPlayerShoot(
			shooterStatus: PlayerStatus
		) {
			if !gameState.ballsAreRolling()
			&& !computerPlayer.shotIsInProgress()
			&& !shooterStatus.gameIsOver() {
		
				//	If the computer player has the cue ball inHand,
				//	place it on the table.
				if gameState.balls[0].status == .inHand {
					gameState.inHandCueBallPose = nil
					gameState.balls[0].placement = validCueBallPlacementNearCenter(in: gameState)
					gameState.balls[0].status = .inPlay
					Task { await NEBSoundPlayer.shared.play(.ballOnTable, volume: 1.0) }
				}

				//	Let the computer player initiate the shot.
				computerPlayer.planShot(
								in: gameState,
								shooterStatus: shooterStatus,
								difficulty: computerPlayerDifficulty,
								startTime: thePredictedPresentationTime)
			}
		}
		
		//	Is it time for our computer opponent to initiate a shot?
		if gameState.whoseTurn == .computer {
			maybeLetComputerPlayerShoot(shooterStatus: gameState.playerStatus[.computer]!)
		}

#if targetEnvironment(simulator)
		//	There's no hand-tracking in the simulator and therefore
		//	no way to take a shot. So when running in the simulator,
		//	let's let the computer play for "self".
		//	Is it time for our computer self to initiate a shot?
		if gameState.whoseTurn == .human {
			maybeLetComputerPlayerShoot(shooterStatus: gameState.playerStatus[.human]!)
		}
#endif

		//	Advance the animation for one frame interval.
		_ = gameState.update(for: theFrameInterval)
		
		theFrame.endUpdate()

// MARK: -
// MARK: Get drawable

		//	Wait until the optimal time for querying
		//	the headset and hand positions.
		LayerRenderer.Clock().wait(until: theTiming.optimalInputTime)

		//	Ask for theDrawable here, before calling startSubmission(),
		//	to avoid the risk of either not calling endSubmission(),
		//	or calling it without having submitted a frame
		//	(which would generate a run-time error).
		//
		//		Note: Once we get a drawable, we're committed
		//		to submitting a frame. Otherwise we get an error
		//		message about "more than 3 frames in flight".
		//
		guard let theDrawable = theFrame.queryDrawable()
		else {
			//	When dismissing the immersive space, we may fail
			//	to get theDrawable on our last pass through the render loop.
			//	This is OK.
			return
		}
		
// MARK: -
// MARK: "Submission" section

		//	Query the headset and hand positions
		//	and submit calls to render the frame.
		//	The startSubmission() / endSubmission() calls help
		//	Compositor Services plan its timing for future frames.

		theFrame.startSubmission()

		let thePresentationTime = LayerRenderer.Clock.Instant.epoch.duration(to:
			theDrawable.frameTiming.presentationTime).timeInterval

		let theDeviceAnchor = worldTrackingProvider.queryDeviceAnchor(
									atTimestamp: thePresentationTime)

		let thePhysicalHeadsetPoseInWorldCoords = convertFloat4x4toDouble4x4(
			theDeviceAnchor?.originFromAnchorTransform ?? matrix_identity_float4x4)
		var thePhysicalHeadsetPoseInTableCoords = thePhysicalHeadsetPoseInWorldCoords
		thePhysicalHeadsetPoseInTableCoords[3][1] -= tableHeight
		let theVirtualHeadsetPose = poseMapper.getVirtualHeadsetPose(
								physicalPose: thePhysicalHeadsetPoseInTableCoords)

		//	latestAnchors gives a *much* smoother animation than handAnchors(at:)
		let theHandAnchors = handTrackingProvider.latestAnchors

		//	Convert the joints' poses to virtual table coordinates.
		let theJointPoses: [PlayersHand : VirtualJointPoses] = PlayersHand.allCases.reduce(into: [:]) { jointPosesDictionary, playersHand in

			let theHandAnchor: HandAnchor?
			switch playersHand {
			case .left:  theHandAnchor = theHandAnchors.leftHand
			case .right: theHandAnchor = theHandAnchors.rightHand
			}
			
			let theJointPoses
				=  VirtualJointPoses(
					for: theHandAnchor,
					spaceGeometry: gameState.spaceGeometry,
					physicalHeadsetPose: thePhysicalHeadsetPoseInWorldCoords,
					virtualHeadsetPose: theVirtualHeadsetPose)	//	in table coords
			
			jointPosesDictionary[playersHand] = theJointPoses
		}

		let theVirtualHandVertexArrays: [PlayersHand: [NEBVertex]]
			= theJointPoses.mapValues() { virtualJointPoses in
		
			let theHandVertexPositions
				= handMeshVertices(virtualJointPoses: virtualJointPoses)
			
			let theHandVertices = theHandVertexPositions.map() { v in
				NEBVertex(
					pos: SIMD4<Float32>(v),
					tex: .zero)
			}
			
			return theHandVertices
		}

		let theGripHandJointPoses: VirtualJointPoses?	//	player's rear hand
		let theBridgeHandJointPoses: VirtualJointPoses?	//	player's forward hand
		switch playerHandedness {
		
		case .leftHanded:
			theGripHandJointPoses   = theJointPoses[.left]
			theBridgeHandJointPoses = theJointPoses[.right]

		case .rightHanded:
			theGripHandJointPoses   = theJointPoses[.right]
			theBridgeHandJointPoses = theJointPoses[.left]
		}

		//	Pick the two points we'd like the cue stick to pass through,
		//	in virtual table coordinates.
		let theGripPoint   = theGripHandJointPoses?.cueStickGripPoint()
		let theBridgePoint = theBridgeHandJointPoses?.cueStickBridgePoint()

		updateGameStartVariables(
			physicalHeadsetPoseInTableCoordinates: thePhysicalHeadsetPoseInTableCoords,
			bridgePointInVirtualCoordinates: theBridgePoint)

#if targetEnvironment(simulator)
		let theOwnCueStickPose = gameState.whoseTurn == .human ?
			computerPlayer.cueStickPose(atTime: thePresentationTime) : nil
#else
		let theOwnCueStickPose
			= cueStickIsAvailable ?
				getVirtualCueStickPose(
					gripPoint: theGripPoint,
					bridgePoint: theBridgePoint,
					cueBallIsInHand: gameState.balls[0].status == .inHand,
					spaceGeometry: gameState.spaceGeometry) :
				nil
#endif
		gameState.cueStick[.human]!.movedTo(theOwnCueStickPose)

		let theComputersCueStickPose: simd_double4x4?
		switch gameState.whoseTurn {
		case .human:
			theComputersCueStickPose = nil
		case .computer:
			theComputersCueStickPose = computerPlayer.cueStickPose(atTime: thePresentationTime)
		}

		gameState.cueStick[.computer]!.movedTo(theComputersCueStickPose)
		
		if gameState.balls[0].status == .inHand {
		
			switch gameState.whoseTurn {
			
			case .human:
			
				let theCueBallCenterPoint = theBridgeHandJointPoses?.cueBallInHandPoint()
			
				gameState.inHandCueBallPose
					= cueBallIsAvailable ?
						getVirtualInHandCueBallPose(
							cueBallCenterPoint: theCueBallCenterPoint,
							spaceGeometry: gameState.spaceGeometry) :
					  	nil
				
				if playerHasRaisedCueBallAboveTable
				&& gameState.maybePlaceCueBall()	//	This is where the cue ball gets placed.
				&& !playerHasPlacedCueBallOnTable {
					
					playerHasPlacedCueBallOnTable = true
					
					//	Let the Control Panel know the player
					//	has placed the cue ball on the table.
					playerPlacedCueBall()
				}
			
			case .computer:
			
				let theHeightAboveTable = 0.35	//	in meters

				gameState.inHandCueBallPose = axisAlignedTranslation(
					axis: 1,
					distanceInRadians: theHeightAboveTable / gameState.spaceGeometry.spaceRadius,
					geometryType: gameState.spaceGeometry.geometryType)
			}
		}

		//	On the one hand, Apple's web page
		//		https://developer.apple.com/documentation/compositorservices/layerrenderer/drawable/4295946-deviceanchor
		//	says
		//		At display time, Compositor Services compares your predicted
		//		device anchor with the actual position and orientation of the device
		//		at that time. If the two values don’t match, Compositor Services
		//		adjusts the pixels of your frame to align it with the hardware.
		//		If you don’t want it to make this adjustment, set this property to nil.
		//
		//	On the other hand, the documentation in the source code for
		//		LayerRenderer.Drawable.deviceAnchor
		//	says
		//		[deviceAnchor] is used by the compositor to reproject
		//		a drawable when no new drawables are available.
		//	where "reproject" means to re-present the frame but
		//	shifted by a few pixels according to where the user's head is.
		//
		//	In practice, if one sets theDrawable.deviceAnchor = nil
		//	one gets an error message
		//
		//		Presenting a drawable without a device anchor.
		//		This drawable won't be presented.
		//
		//	so it seems that the web page is wrong and
		//	the source code documentation is correct.
		//
		theDrawable.deviceAnchor = theDeviceAnchor

		guard let theCommandBuffer = commandQueue.makeCommandBuffer() else {
			preconditionFailure("Failed to make theCommandBuffer")
		}
		renderTheFrame(
			drawable: theDrawable,
			virtualHeadsetPose: theVirtualHeadsetPose,
			virtualHandVertexArrays: theVirtualHandVertexArrays,
			commandBuffer: theCommandBuffer)
		theDrawable.encodePresent(commandBuffer: theCommandBuffer)
		theCommandBuffer.commit()
		
		theFrame.endSubmission()

#warning("Delete the following code in some future release, when players are used to pressing the Digital Crown to close the immersive space.")
		var thePlayersHandIsRaised = false
		for thePlayersHand in PlayersHand.allCases {
			let theOptionalHandAnchor: HandAnchor?
			switch thePlayersHand {
			case .left:  theOptionalHandAnchor = theHandAnchors.leftHand
			case .right: theOptionalHandAnchor = theHandAnchors.rightHand
			}
			guard let theHandAnchor = theOptionalHandAnchor,
				  theHandAnchor.isTracked
			else { continue }
			let thePhysicalHandPoseInWorldCoords = convertFloat4x4toDouble4x4(
								theHandAnchor.originFromAnchorTransform)
			if thePhysicalHandPoseInWorldCoords[3][1]
			 > thePhysicalHeadsetPoseInWorldCoords[3][1] {
				thePlayersHandIsRaised = true
			}
		}
		let thePlayerJustNowRaisedHand
			= thePlayersHandIsRaised && !playersHandWasAlreadyRaised
		playersHandWasAlreadyRaised = thePlayersHandIsRaised
		if !instructionsPanelIsVisible
		&& thePlayerJustNowRaisedHand {
			temporaryClosure()
		}
	}
	
	func updateGameStartVariables(
		physicalHeadsetPoseInTableCoordinates: simd_double4x4,
		bridgePointInVirtualCoordinates: SIMD4<Double>?	//	in virtual table coordinates
	) {
	
		//	The player will be clear of the table when the absolute value
		//	of the headset's x or z coordinate exceeds half the table width.
		//	But let's test against 0.4 times the table width instead,
		//	so that if the player is standing right at the table edge
		//	leaning forward, we'll accept that as "seeing the whole table".
		let theThreshold = 0.4 * gameState.spaceGeometry.tableEdgeLengthInMeters
	
		//	Once we set these variables to true,
		//	we never set them back to false.
		
		if !playerHasSteppedBackFromTable {
		
			playerHasSteppedBackFromTable =
				abs(physicalHeadsetPoseInTableCoordinates[3][0]) > theThreshold
			 || abs(physicalHeadsetPoseInTableCoordinates[3][2]) > theThreshold
			
			if playerHasSteppedBackFromTable {
			
				//	Give the player a cue ball to place on the table.
				cueBallIsAvailable = true
				
				//	Let the Control Panel know the player has stepped back.
				playerSteppedBack()
			}
		}
		
		//	Wait until the player has stepped back from the table
		//	before checking whether s/he has raised the cue ball.

		if playerHasSteppedBackFromTable
		&& !playerHasRaisedCueBallAboveTable {
		
			if let theBridgePoint = bridgePointInVirtualCoordinates {
			
				playerHasRaisedCueBallAboveTable =
					theBridgePoint.y > 3.0 * ballRadius
				
				if playerHasRaisedCueBallAboveTable {
					//	Let the Control Panel know the player
					//	has raised the cue ball above the table.
					playerRaisedCueBall()
				}
			}
		}

		//	We'll set playerHasPlacedCueBallOnTable
		//	when gameState.maybePlaceCueBall() returns true.
	}

	func renderTheFrame(
		drawable: LayerRenderer.Drawable,
		virtualHeadsetPose: simd_double4x4,
		virtualHandVertexArrays: [PlayersHand: [NEBVertex]],
		commandBuffer: MTLCommandBuffer
	) {

		let theRenderPassDescriptor = ggvrRenderPassDescriptor(
										layerRenderer: layerRenderer,
										drawable: drawable)

		//	Get an available view-projection buffer.
		//
		//		The .get() method automatically creates and installs
		//		a completedHandler to put the buffer back into the pool
		//		when the GPU has finished executing the commandBuffer.
		//
		let theViewProjectionBuffer = viewProjectionBufferPool.get(forUseWith: commandBuffer)

		ggvrPrepareViewProjectionBuffer(
			viewProjectionBuffer: theViewProjectionBuffer,
			virtualHeadsetPose: virtualHeadsetPose,
			drawable: drawable,
			spaceGeometry: gameState.spaceGeometry,
			adjustPerceivedDistances: adjustPerceivedDistances)

		if let theRenderEncoder = commandBuffer.makeRenderCommandEncoder(
										descriptor: theRenderPassDescriptor) {
			theRenderEncoder.label = "Non-Euclidean Billiards render encoder"

			ggvrEncodeViewports(
				drawable: drawable,
				renderEncoder: theRenderEncoder)

			encodeCommands(
				renderEncoder: theRenderEncoder,
				viewProjectionBuffer: theViewProjectionBuffer,
				virtualHandVertexArrays: virtualHandVertexArrays)
				
			theRenderEncoder.endEncoding()
		}
	}
	
	func encodeCommands(
		renderEncoder: MTLRenderCommandEncoder,
		viewProjectionBuffer: MTLBuffer,
		virtualHandVertexArrays: [PlayersHand: [NEBVertex]]
	) {
		let theMatrixSize = MemoryLayout<simd_float4x4>.stride

		renderEncoder.pushDebugGroup("Draw the whole scene")

		renderEncoder.setCullMode(.back)
		renderEncoder.setFrontFacing(.clockwise)	//	the default
		renderEncoder.setVertexBuffer(
						viewProjectionBuffer,
						offset: 0,
						index: BufferIndexViewProjection)
				
		//	Table

		renderEncoder.setRenderPipelineState(billiardsTablePipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						tableMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		var theTablePose = matrix_identity_float4x4	//	table pose in table coordinates
		renderEncoder.setVertexBytes(
			&theTablePose,
			length: theMatrixSize,
			index: BufferIndexModelMatrix)

		renderEncoder.drawIndexedPrimitives(
						type: .triangle,
						indexCount: tableMesh.indexCount,
						indexType: .uint16,
						indexBuffer: tableMesh.indexBuffer,
						indexBufferOffset: 0,
						instanceCount: 1)
		
		//	Grounding shadows

		renderEncoder.setRenderPipelineState(groundingShadowPipelineState)
		renderEncoder.setDepthStencilState(disabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						groundingShadowMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		for i in 0 ..< numBalls {
		
			let theBall = gameState.balls[i]

			if theBall.status == .inPlay {
		
				var theShadowPose = convertDouble4x4toFloat4x4(theBall.placement)
				renderEncoder.setVertexBytes(
					&theShadowPose,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: groundingShadowMesh.indexCount,
								indexType: .uint16,
								indexBuffer: groundingShadowMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}
		
		//	Pockets
		
		renderEncoder.setRenderPipelineState(pocketPipelineState)
		renderEncoder.setDepthStencilState(disabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						pocketMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)

		for thePose in gameState.spaceGeometry.pocketPoses {
		
			var thePocketPose = convertDouble4x4toFloat4x4(thePose)
			renderEncoder.setVertexBytes(
				&thePocketPose,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: pocketMesh.indexCount,
							indexType: .uint16,
							indexBuffer: pocketMesh.indexBuffer,
							indexBufferOffset: 0,
							instanceCount: 1)
		}
		
		//	Balls
		
		let theWorldUpVector = SIMD4<Double>(0.0, 1.0, 0.0, 0.0)

		renderEncoder.setRenderPipelineState(billiardBallPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						ballMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)
		renderEncoder.setFragmentSamplerState(
						ballSampler,
						index: SamplerIndexPrimary)
			
		for i in 0 ..< numBalls {

			let theBall = gameState.balls[i]

			if theBall.status == .inPlay {

				var theBallPose = convertDouble4x4toFloat4x4(
									theBall.placement	//	right-to-left matrix action
								  * gameState.spaceGeometry.ballLift
								  * theBall.rotation)
				renderEncoder.setVertexBytes(
					&theBallPose,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)
				
				let theUpVector4 = geometricInverse(
									of: theBall.rotation,
									geometryType: gameState.spaceGeometry.geometryType)
								 * theWorldUpVector
				var theUpVector3 = SIMD3<Float16>(
										Float16(theUpVector4.x),
										Float16(theUpVector4.y),
										Float16(theUpVector4.z))
				renderEncoder.setVertexBytes(
					&theUpVector3,
					length: MemoryLayout<SIMD3<Float16>>.stride,
					index: BufferIndexUpVector)

				var theColorSaturation = Float16(
					colorTargetBallsOnly ? theBall.colorSaturation : 1.0)
				renderEncoder.setFragmentBytes(
					&theColorSaturation,
					length: MemoryLayout<Float16>.stride,
					index: BufferIndexColoration)

				renderEncoder.setFragmentTexture(
								ballTextures[i],
								index: TextureIndexPrimary)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: ballMesh.indexCount,
								indexType: .uint16,
								indexBuffer: ballMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}

		if let theInHandCueBallPose = gameState.inHandCueBallPose {	//	in table coordinates
		
			var theInHandPose = convertDouble4x4toFloat4x4(theInHandCueBallPose)
			renderEncoder.setVertexBytes(
				&theInHandPose,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			//	The inHandCueBallPose is a plain translation, so its
			//	local up-vector is a reasonable global up-vector.
			var theUpVector = SIMD3<Float16>(0.0, 1.0, 0.0)
			renderEncoder.setVertexBytes(
				&theUpVector,
				length: MemoryLayout<SIMD3<Float16>>.stride,
				index: BufferIndexUpVector)

			var theColorSaturation = Float16(1.0)
			renderEncoder.setFragmentBytes(
				&theColorSaturation,
				length: MemoryLayout<Float16>.stride,
				index: BufferIndexColoration)

			renderEncoder.setFragmentTexture(
							ballTextures[0],
							index: TextureIndexPrimary)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: ballMesh.indexCount,
							indexType: .uint16,
							indexBuffer: ballMesh.indexBuffer,
							indexBufferOffset: 0,
							instanceCount: 1)
		}
		
		//	Cue sticks
		
		renderEncoder.setRenderPipelineState(cueStickPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)
		renderEncoder.setVertexBuffer(
						cueStickMesh.vertexBuffer,
						offset: 0,
						index: BufferIndexVertexAttributes)
		renderEncoder.setFragmentTexture(
						cueStickTexture,
						index: TextureIndexPrimary)
		renderEncoder.setFragmentSamplerState(
						cueStickSampler,
						index: SamplerIndexPrimary)

		func renderCueStick(
			_ cueStick: CueStick
		) {
			if let theCueStickPose = cueStick.current,
			   cueStick.opacity > 0.0 {

				var theCueStickPoseFloat = convertDouble4x4toFloat4x4(theCueStickPose)
				renderEncoder.setVertexBytes(
					&theCueStickPoseFloat,
					length: theMatrixSize,
					index: BufferIndexModelMatrix)

				var theOpacity = Float16(cueStick.opacity)
				renderEncoder.setFragmentBytes(
					&theOpacity,
					length: MemoryLayout<Float16>.stride,
					index: BufferIndexColoration)

				renderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexCount: cueStickMesh.indexCount,
								indexType: .uint16,
								indexBuffer: cueStickMesh.indexBuffer,
								indexBufferOffset: 0,
								instanceCount: 1)
			}
		}
		renderCueStick(gameState.cueStick[.human]!)
		renderCueStick(gameState.cueStick[.computer]!)
	
		//	Hands
			
		//	To keep the triangle count low, the player's hands
		//	are 2D surfaces, not 3D models.

		renderEncoder.setRenderPipelineState(handPipelineState)
		renderEncoder.setDepthStencilState(enabledDepthStencilState)

		for thePlayersHand in PlayersHand.allCases {

			guard let theVirtualHandVertexArray = virtualHandVertexArrays[thePlayersHand]
			else { continue }

			//	The hand vertices are defined in the virtual world space,
			//	so the hand pose is always the identity.
			var theIdentityMatrix = matrix_identity_float4x4
			renderEncoder.setVertexBytes(
				&theIdentityMatrix,
				length: theMatrixSize,
				index: BufferIndexModelMatrix)

			renderEncoder.setVertexBytes(
				theVirtualHandVertexArray,
				length: theVirtualHandVertexArray.count * MemoryLayout<NEBVertex>.stride,
				index: BufferIndexVertexAttributes)
			
			//	Render back of hand
			
			//	The left and right hand meshes are topologically
			//	mirror images of each other.
			switch thePlayersHand {
			case .left:  renderEncoder.setCullMode(.front)
			case .right: renderEncoder.setCullMode(.back)
			}

			//	A selection of reasonable skin tones
			//	in linear Display P3:
			//
			//		SIMD3<Float16>(0.2351, 0.0966, 0.0271)
			//		SIMD3<Float16>(0.5067, 0.2492, 0.0765)
			//		SIMD3<Float16>(0.6863, 0.4235, 0.1712)
			//		SIMD3<Float16>(0.8192, 0.5507, 0.2408)
			//		SIMD3<Float16>(0.9482, 0.7180, 0.4439)
			//
			var theBackColor = SIMD4<Float16>(0.59, 0.33, 0.12, 1.0)
			renderEncoder.setFragmentBytes(
				&theBackColor,
				length: MemoryLayout<SIMD4<Float16>>.stride,
				index: BufferIndexColoration)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: gHandMeshFacets.count,
							indexType: .uint16,
							indexBuffer: handFacets,
							indexBufferOffset: 0)
			
			//	Render palm of hand
			
			//	Draw the faces that got culled when drawing
			//	the back of the hand (above).
			switch thePlayersHand {
			case .left:  renderEncoder.setCullMode(.back)
			case .right: renderEncoder.setCullMode(.front)
			}

			//	See skin tones above.
			var thePalmColor = SIMD4<Float16>(0.82, 0.55, 0.24, 1.0)
			renderEncoder.setFragmentBytes(
				&thePalmColor,
				length: MemoryLayout<SIMD4<Float16>>.stride,
				index: BufferIndexColoration)

			renderEncoder.drawIndexedPrimitives(
							type: .triangle,
							indexCount: gHandMeshFacets.count,
							indexType: .uint16,
							indexBuffer: handFacets,
							indexBufferOffset: 0)
		}
		renderEncoder.setCullMode(.back)	//	restore our default

		renderEncoder.popDebugGroup()
	}
}
